# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

class Permissions:
    """Wraps up the Discord permission value.

    Supported operations:

    +-----------+------------------------------------------+
    | Operation |               Description                |
    +===========+==========================================+
    | x == y    | Checks if two permissions are equal.     |
    +-----------+------------------------------------------+
    | x != y    | Checks if two permissions are not equal. |
    +-----------+------------------------------------------+
    | x <= y    | Checks if a permission is a subset       |
    |           | of another permission.                   |
    +-----------+------------------------------------------+
    | x >= y    | Checks if a permission is a superset     |
    |           | of another permission.                   |
    +-----------+------------------------------------------+
    | x < y     | Checks if a permission is a strict       |
    |           | subset of another permission.            |
    +-----------+------------------------------------------+
    | x > y     | Checks if a permission is a strict       |
    |           | superset of another permission.          |
    +-----------+------------------------------------------+
    | hash(x)   | Return the permission's hash.            |
    +-----------+------------------------------------------+
    | iter(x)   | Returns an iterator of (perm, value)     |
    |           | pairs. This allows this class to be used |
    |           | as an iterable in e.g. set/list/dict     |
    |           | constructions.                           |
    +-----------+------------------------------------------+

    The properties provided are two way. You can set and retrieve individual bits using the properties as if they
    were regular bools. This allows you to edit permissions.

    Attributes
    -----------
    value
        The raw value. This value is a bit array field of a 32-bit integer
        representing the currently available permissions. You should query
        permissions via the properties rather than using this raw value.
    """

    __slots__ = [ 'value' ]
    def __init__(self, permissions=0, **kwargs):
        self.value = permissions

    def __eq__(self, other):
        return isinstance(other, Permissions) and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.value)

    def _perm_iterator(self):
        for attr in dir(self):
            # check if it's a property, because if so it's a permission
            is_property = isinstance(getattr(self.__class__, attr), property)
            if is_property:
                yield (attr, getattr(self, attr))

    def __iter__(self):
        return self._perm_iterator()

    def is_subset(self, other):
        """Returns True if other has the same or fewer permissions as self."""
        if isinstance(other, Permissions):
            return (self.value & other.value) == self.value
        else:
            raise TypeError("cannot compare {} with {}".format(self.__class__.__name__, other.__class__name))

    def is_superset(self, other):
        """Returns True if other has the same or more permissions as self."""
        if isinstance(other, Permissions):
            return (self.value | other.value) == self.value
        else:
            raise TypeError("cannot compare {} with {}".format(self.__class__.__name__, other.__class__name))

    def is_strict_subset(self, other):
        """Returns True if the permissions on other are a strict subset of those on self."""
        return self.is_subset(other) and self != other

    def is_strict_superset(self, other):
        """Returns True if the permissions on other are a strict superset of those on self."""
        return self.is_superset(other) and self != other

    __le__ = is_subset
    __ge__ = is_superset
    __lt__ = is_strict_subset
    __gt__ = is_strict_superset

    @classmethod
    def none(cls):
        """A factory method that creates a :class:`Permissions` with all
        permissions set to False."""
        return cls(0)

    @classmethod
    def all(cls):
        """A factory method that creates a :class:`Permissions` with all
        permissions set to True."""
        return cls(0b01111111111101111111110001111111)

    @classmethod
    def all_channel(cls):
        """A :class:`Permissions` with all channel-specific permissions set to
        True and the server-specific ones set to False. The server-specific
        permissions are currently:

        - manager_server
        - kick_members
        - ban_members
        - administrator
        - change_nicknames
        - manage_nicknames
        """
        return cls(0b00110011111101111111110001010001)

    @classmethod
    def general(cls):
        """A factory method that creates a :class:`Permissions` with all
        "General" permissions from the official Discord UI set to True."""
        return cls(0b01111100000000000000000000111111)

    @classmethod
    def text(cls):
        """A factory method that creates a :class:`Permissions` with all
        "Text" permissions from the official Discord UI set to True."""
        return cls(0b00000000000001111111110001000000)

    @classmethod
    def voice(cls):
        """A factory method that creates a :class:`Permissions` with all
        "Voice" permissions from the official Discord UI set to True."""
        return cls(0b00000011111100000000000000000000)

    def update(self, **kwargs):
        """Bulk updates this permission object.

        Allows you to set multiple attributes by using keyword
        arguments. The names must be equivalent to the properties
        listed. Extraneous key/value pairs will be silently ignored.

        Parameters
        ------------
        \*\*kwargs
            A list of key/value pairs to bulk update permissions with.
        """
        for key, value in kwargs.items():
            try:
                is_property = isinstance(getattr(self.__class__, key), property)
            except AttributeError:
                continue

            if is_property:
                setattr(self, key, value)

    def _bit(self, index):
        return bool((self.value >> index) & 1)

    def _set(self, index, value):
        if value == True:
            self.value |= (1 << index)
        elif value == False:
            self.value &= ~(1 << index)
        else:
            raise TypeError('Value to set for Permissions must be a bool.')

    def handle_overwrite(self, allow, deny):
        # Basically this is what's happening here.
        # We have an original bit array, e.g. 1010
        # Then we have another bit array that is 'denied', e.g. 1111
        # And then we have the last one which is 'allowed', e.g. 0101
        # We want original OP denied to end up resulting in
        # whatever is in denied to be set to 0.
        # So 1010 OP 1111 -> 0000
        # Then we take this value and look at the allowed values.
        # And whatever is allowed is set to 1.
        # So 0000 OP2 0101 -> 0101
        # The OP is base  & ~denied.
        # The OP2 is base | allowed.
        self.value = (self.value & ~deny) | allow

    @property
    def create_instant_invite(self):
        """Returns True if the user can create instant invites."""
        return self._bit(0)

    @create_instant_invite.setter
    def create_instant_invite(self, value):
        self._set(0, value)

    @property
    def kick_members(self):
        """Returns True if the user can kick users from the server."""
        return self._bit(1)

    @kick_members.setter
    def kick_members(self, value):
        self._set(1, value)

    @property
    def ban_members(self):
        """Returns True if a user can ban users from the server."""
        return self._bit(2)

    @ban_members.setter
    def ban_members(self, value):
        self._set(2, value)

    @property
    def administrator(self):
        """Returns True if a user is an administrator. This role overrides all other permissions.

        This also bypasses all channel-specific overrides.
        """
        return self._bit(3)

    @administrator.setter
    def administrator(self, value):
        self._set(3, value)

    @property
    def manage_channels(self):
        """Returns True if a user can edit, delete, or create channels in the server.

        This also corresponds to the "manage channel" channel-specific override."""
        return self._bit(4)

    @manage_channels.setter
    def manage_channels(self, value):
        self._set(4, value)

    @property
    def manage_server(self):
        """Returns True if a user can edit server properties."""
        return self._bit(5)

    @manage_server.setter
    def manage_server(self, value):
        self._set(5, value)

    @property
    def add_reactions(self):
        """Returns True if a user can add reactions to messages."""
        return self._bit(6)

    @add_reactions.setter
    def add_reactions(self, value):
        self._set(6, value)

    # 4 unused

    @property
    def read_messages(self):
        """Returns True if a user can read messages from all or specific text channels."""
        return self._bit(10)

    @read_messages.setter
    def read_messages(self, value):
        self._set(10, value)

    @property
    def send_messages(self):
        """Returns True if a user can send messages from all or specific text channels."""
        return self._bit(11)

    @send_messages.setter
    def send_messages(self, value):
        self._set(11, value)

    @property
    def send_tts_messages(self):
        """Returns True if a user can send TTS messages from all or specific text channels."""
        return self._bit(12)

    @send_tts_messages.setter
    def send_tts_messages(self, value):
        self._set(12, value)

    @property
    def manage_messages(self):
        """Returns True if a user can delete messages from a text channel. Note that there are currently no ways to edit other people's messages."""
        return self._bit(13)

    @manage_messages.setter
    def manage_messages(self, value):
        self._set(13, value)

    @property
    def embed_links(self):
        """Returns True if a user's messages will automatically be embedded by Discord."""
        return self._bit(14)

    @embed_links.setter
    def embed_links(self, value):
        self._set(14, value)

    @property
    def attach_files(self):
        """Returns True if a user can send files in their messages."""
        return self._bit(15)

    @attach_files.setter
    def attach_files(self, value):
        self._set(15, value)

    @property
    def read_message_history(self):
        """Returns True if a user can read a text channel's previous messages."""
        return self._bit(16)

    @read_message_history.setter
    def read_message_history(self, value):
        self._set(16, value)

    @property
    def mention_everyone(self):
        """Returns True if a user's @everyone will mention everyone in the text channel."""
        return self._bit(17)

    @mention_everyone.setter
    def mention_everyone(self, value):
        self._set(17, value)

    @property
    def external_emojis(self):
        """Returns True if a user can use emojis from other servers."""
        return self._bit(18)

    @external_emojis.setter
    def external_emojis(self, value):
        self._set(18, value)

    # 1 unused

    @property
    def connect(self):
        """Returns True if a user can connect to a voice channel."""
        return self._bit(20)

    @connect.setter
    def connect(self, value):
        self._set(20, value)

    @property
    def speak(self):
        """Returns True if a user can speak in a voice channel."""
        return self._bit(21)

    @speak.setter
    def speak(self, value):
        self._set(21, value)

    @property
    def mute_members(self):
        """Returns True if a user can mute other users."""
        return self._bit(22)

    @mute_members.setter
    def mute_members(self, value):
        self._set(22, value)

    @property
    def deafen_members(self):
        """Returns True if a user can deafen other users."""
        return self._bit(23)

    @deafen_members.setter
    def deafen_members(self, value):
        self._set(23, value)

    @property
    def move_members(self):
        """Returns True if a user can move users between other voice channels."""
        return self._bit(24)

    @move_members.setter
    def move_members(self, value):
        self._set(24, value)

    @property
    def use_voice_activation(self):
        """Returns True if a user can use voice activation in voice channels."""
        return self._bit(25)

    @use_voice_activation.setter
    def use_voice_activation(self, value):
        self._set(25, value)

    @property
    def change_nickname(self):
        """Returns True if a user can change their nickname in the server."""
        return self._bit(26)

    @change_nickname.setter
    def change_nickname(self, value):
        self._set(26, value)

    @property
    def manage_nicknames(self):
        """Returns True if a user can change other user's nickname in the server."""
        return self._bit(27)

    @manage_nicknames.setter
    def manage_nicknames(self, value):
        self._set(27, value)

    @property
    def manage_roles(self):
        """Returns True if a user can create or edit roles less than their role's position.

        This also corresponds to the "manage permissions" channel-specific override.
        """
        return self._bit(28)

    @manage_roles.setter
    def manage_roles(self, value):
        self._set(28, value)

    @property
    def manage_webhooks(self):
        """Returns True if a user can create, edit, or delete webhooks."""
        return self._bit(29)

    @manage_webhooks.setter
    def manage_webhooks(self, value):
        self._set(29, value)

    @property
    def manage_emojis(self):
        """Returns True if a user can create, edit, or delete emojis."""
        return self._bit(30)

    @manage_emojis.setter
    def manage_emojis(self, value):
        self._set(30, value)

    # 1 unused

    # after these 32 bits, there's 21 more unused ones technically

def augment_from_permissions(cls):
    cls.VALID_NAMES = { name for name in dir(Permissions) if isinstance(getattr(Permissions, name), property) }

    # make descriptors for all the valid names
    for name in cls.VALID_NAMES:
        # god bless Python
        def getter(self, x=name):
            return self._values.get(x)
        def setter(self, value, x=name):
            self._set(x, value)

        prop = property(getter, setter)
        setattr(cls, name, prop)

    return cls

@augment_from_permissions
class PermissionOverwrite:
    """A type that is used to represent a channel specific permission.

    Unlike a regular :class:`Permissions`\, the default value of a
    permission is equivalent to ``None`` and not ``False``. Setting
    a value to ``False`` is **explicitly** denying that permission,
    while setting a value to ``True`` is **explicitly** allowing
    that permission.

    The values supported by this are the same as :class:`Permissions`
    with the added possibility of it being set to ``None``.

    Supported operations:

    +-----------+------------------------------------------+
    | Operation |               Description                |
    +===========+==========================================+
    | iter(x)   | Returns an iterator of (perm, value)     |
    |           | pairs. This allows this class to be used |
    |           | as an iterable in e.g. set/list/dict     |
    |           | constructions.                           |
    +-----------+------------------------------------------+

    Parameters
    -----------
    \*\*kwargs
        Set the value of permissions by their name.
    """

    def __init__(self, **kwargs):
        self._values = {}

        for key, value in kwargs.items():
            setattr(self, key, value)

    def _set(self, key, value):
        if value not in (True, None, False):
            raise TypeError('Expected bool or NoneType, received {0.__class__.__name__}'.format(value))

        self._values[key] = value

    def pair(self):
        """Returns the (allow, deny) pair from this overwrite.

        The value of these pairs is :class:`Permissions`.
        """

        allow = Permissions.none()
        deny  = Permissions.none()

        for key, value in self._values.items():
            if value is True:
                setattr(allow, key, True)
            elif value is False:
                setattr(deny, key, True)

        return allow, deny

    @classmethod
    def from_pair(cls, allow, deny):
        """Creates an overwrite from an allow/deny pair of :class:`Permissions`."""
        ret = cls()
        for key, value in allow:
            if value is True:
                setattr(ret, key, True)

        for key, value in deny:
            if value is True:
                setattr(ret, key, False)

        return ret

    def is_empty(self):
        """Checks if the permission overwrite is currently empty.

        An empty permission overwrite is one that has no overwrites set
        to True or False.
        """
        return all(x is None for x in self._values.values())

    def update(self, **kwargs):
        """Bulk updates this permission overwrite object.

        Allows you to set multiple attributes by using keyword
        arguments. The names must be equivalent to the properties
        listed. Extraneous key/value pairs will be silently ignored.

        Parameters
        ------------
        \*\*kwargs
            A list of key/value pairs to bulk update with.
        """
        for key, value in kwargs.items():
            if key not in self.VALID_NAMES:
                continue

            setattr(self, key, value)

    def __iter__(self):
        for key in self.VALID_NAMES:
            yield key, self._values.get(key)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import aiohttp
import asyncio
import json
import sys
import logging
import inspect
import weakref
from random import randint as random_integer

log = logging.getLogger(__name__)

from .errors import HTTPException, Forbidden, NotFound, LoginFailure, GatewayNotFound
from . import utils, __version__

@asyncio.coroutine
def json_or_text(response):
    text = yield from response.text(encoding='utf-8')
    if response.headers['content-type'] == 'application/json':
        return json.loads(text)
    return text

def _func_():
    # emulate __func__ from C++
    return inspect.currentframe().f_back.f_code.co_name

class HTTPClient:
    """Represents an HTTP client sending HTTP requests to the Discord API."""

    BASE          = 'https://discordapp.com'
    API_BASE      = BASE     + '/api/v6'
    GATEWAY       = API_BASE + '/gateway'
    USERS         = API_BASE + '/users'
    ME            = USERS    + '/@me'
    REGISTER      = API_BASE + '/auth/register'
    LOGIN         = API_BASE + '/auth/login'
    LOGOUT        = API_BASE + '/auth/logout'
    GUILDS        = API_BASE + '/guilds'
    CHANNELS      = API_BASE + '/channels'
    APPLICATIONS  = API_BASE + '/oauth2/applications'

    SUCCESS_LOG = '{method} {url} has received {text}'
    REQUEST_LOG = '{method} {url} with {json} has returned {status}'

    def __init__(self, connector=None, *, loop=None):
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.connector = connector
        self.session = aiohttp.ClientSession(connector=connector, loop=self.loop)
        self._locks = weakref.WeakValueDictionary()
        self.token = None
        self.bot_token = False

        user_agent = 'DiscordBot (https://github.com/Rapptz/discord.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self.user_agent = user_agent.format(__version__, sys.version_info, aiohttp.__version__)

    @asyncio.coroutine
    def request(self, method, url, *, bucket=None, **kwargs):
        lock = self._locks.get(bucket)
        if lock is None:
            lock = asyncio.Lock(loop=self.loop)
            if bucket is not None:
                self._locks[bucket] = lock

        # header creation
        headers = {
            'User-Agent': self.user_agent,
        }

        if self.token is not None:
            headers['Authorization'] = 'Bot ' + self.token if self.bot_token else self.token

        # some checking if it's a JSON request
        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['data'] = utils.to_json(kwargs.pop('json'))

        kwargs['headers'] = headers
        with (yield from lock):
            for tries in range(5):
                r = yield from self.session.request(method, url, **kwargs)
                log.debug(self.REQUEST_LOG.format(method=method, url=url, status=r.status, json=kwargs.get('data')))
                try:
                    # even errors have text involved in them so this is safe to call
                    data = yield from json_or_text(r)

                    # the request was successful so just return the text/json
                    if 300 > r.status >= 200:
                        log.debug(self.SUCCESS_LOG.format(method=method, url=url, text=data))
                        return data

                    # we are being rate limited
                    if r.status == 429:
                        fmt = 'We are being rate limited. Retrying in {:.2} seconds. Handled under the bucket "{}"'

                        # sleep a bit
                        retry_after = data['retry_after'] / 1000.0
                        log.info(fmt.format(retry_after, bucket))
                        yield from asyncio.sleep(retry_after)
                        continue

                    # we've received a 502, unconditional retry
                    if r.status == 502 and tries <= 5:
                        yield from asyncio.sleep(1 + tries * 2)
                        continue

                    # the usual error cases
                    if r.status == 403:
                        raise Forbidden(r, data)
                    elif r.status == 404:
                        raise NotFound(r, data)
                    else:
                        raise HTTPException(r, data)
                finally:
                    # clean-up just in case
                    yield from r.release()

    def get(self, *args, **kwargs):
        return self.request('GET', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.request('PUT', *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.request('PATCH', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.request('DELETE', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request('POST', *args, **kwargs)

    # state management

    @asyncio.coroutine
    def close(self):
        yield from self.session.close()

    def recreate(self):
        self.session = aiohttp.ClientSession(connector=self.connector, loop=self.loop)

    def _token(self, token, *, bot=True):
        self.token = token
        self.bot_token = bot

    # login management

    @asyncio.coroutine
    def email_login(self, email, password):
        payload = {
            'email': email,
            'password': password
        }

        try:
            data = yield from self.post(self.LOGIN, json=payload, bucket=_func_())
        except HTTPException as e:
            if e.response.status == 400:
                raise LoginFailure('Improper credentials have been passed.') from e
            raise

        self._token(data['token'], bot=False)
        return data

    @asyncio.coroutine
    def static_login(self, token, *, bot):
        old_token, old_bot = self.token, self.bot_token
        self._token(token, bot=bot)

        try:
            data = yield from self.get(self.ME)
        except HTTPException as e:
            self._token(old_token, bot=old_bot)
            if e.response.status == 401:
                raise LoginFailure('Improper token has been passed.') from e
            raise e

        return data

    def logout(self):
        return self.post(self.LOGOUT, bucket=_func_())

    # Message management

    def start_private_message(self, user_id):
        payload = {
            'recipient_id': user_id
        }

        return self.post(self.ME + '/channels', json=payload, bucket=_func_())

    def send_message(self, channel_id, content, *, guild_id=None, tts=False, embed=None):
        url = '{0.CHANNELS}/{1}/messages'.format(self, channel_id)
        payload = {
            'nonce': random_integer(-2**63, 2**63 - 1)
        }

        if content:
            payload['content'] = content

        if tts:
            payload['tts'] = True

        if embed:
            payload['embed'] = embed

        return self.post(url, json=payload, bucket='messages:' + str(guild_id))

    def send_typing(self, channel_id):
        url = '{0.CHANNELS}/{1}/typing'.format(self, channel_id)
        return self.post(url, bucket=_func_())

    def send_file(self, channel_id, buffer, *, guild_id=None, filename=None, content=None, tts=False):
        url = '{0.CHANNELS}/{1}/messages'.format(self, channel_id)
        form = aiohttp.FormData()

        if content is not None:
            form.add_field('content', str(content))

        form.add_field('tts', 'true' if tts else 'false')
        form.add_field('file', buffer, filename=filename, content_type='application/octet-stream')

        return self.post(url, data=form, bucket='messages:' + str(guild_id))

    def delete_message(self, channel_id, message_id, guild_id=None):
        url = '{0.CHANNELS}/{1}/messages/{2}'.format(self, channel_id, message_id)
        bucket = '{}:{}'.format(_func_(), guild_id)
        return self.delete(url, bucket=bucket)

    def delete_messages(self, channel_id, message_ids, guild_id=None):
        url = '{0.CHANNELS}/{1}/messages/bulk_delete'.format(self, channel_id)
        payload = {
            'messages': message_ids
        }
        bucket = '{}:{}'.format(_func_(), guild_id)
        return self.post(url, json=payload, bucket=bucket)

    def edit_message(self, message_id, channel_id, content, *, guild_id=None, embed=None):
        url = '{0.CHANNELS}/{1}/messages/{2}'.format(self, channel_id, message_id)
        payload = {}

        if content:
            payload['content'] = content

        if embed:
            payload['embed'] = embed

        return self.patch(url, json=payload, bucket='messages:' + str(guild_id))

    def add_reaction(self, message_id, channel_id, emoji):
        url = '{0.CHANNELS}/{1}/messages/{2}/reactions/{3}/@me'.format(
            self, channel_id, message_id, emoji)
        return self.put(url, bucket='%s:%s' % (_func_(), channel_id))

    def remove_reaction(self, message_id, channel_id, emoji, member_id):
        url = '{0.CHANNELS}/{1}/messages/{2}/reactions/{3}/{4}'.format(
            self, channel_id, message_id, emoji, member_id)
        return self.delete(url, bucket='%s:%s' % (_func_(), channel_id))

    def get_reaction_users(self, message_id, channel_id, emoji, limit, after=None):
        url = '{0.CHANNELS}/{1}/messages/{2}/reactions/{3}'.format(
            self, channel_id, message_id, emoji)
        params = {'limit': limit}
        if after:
            params['after'] = after
        return self.get(url, params=params, bucket='%s:%s' % (_func_(), channel_id))

    def clear_reactions(self, message_id, channel_id):
        url = '{0.CHANNELS}/{1}/messages/{2}/reactions'.format(self, channel_id, message_id)
        return self.delete(url, bucket='%s:%s' % (_func_(), channel_id))

    def get_message(self, channel_id, message_id):
        url = '{0.CHANNELS}/{1}/messages/{2}'.format(self, channel_id, message_id)
        return self.get(url, bucket=_func_())

    def logs_from(self, channel_id, limit, before=None, after=None, around=None):
        url = '{0.CHANNELS}/{1}/messages'.format(self, channel_id)
        params = {
            'limit': limit
        }

        if before:
            params['before'] = before
        if after:
            params['after'] = after
        if around:
            params['around'] = around

        return self.get(url, params=params, bucket=_func_())

    def pin_message(self, channel_id, message_id):
        url = '{0.CHANNELS}/{1}/pins/{2}'.format(self, channel_id, message_id)
        return self.put(url, bucket=_func_())

    def unpin_message(self, channel_id, message_id):
        url = '{0.CHANNELS}/{1}/pins/{2}'.format(self, channel_id, message_id)
        return self.delete(url, bucket=_func_())

    def pins_from(self, channel_id):
        url = '{0.CHANNELS}/{1}/pins'.format(self, channel_id)
        return self.get(url, bucket=_func_())

    # Member management

    def kick(self, user_id, guild_id):
        url = '{0.GUILDS}/{1}/members/{2}'.format(self, guild_id, user_id)
        return self.delete(url, bucket=_func_())

    def ban(self, user_id, guild_id, delete_message_days=1):
        url = '{0.GUILDS}/{1}/bans/{2}'.format(self, guild_id, user_id)
        params = {
            'delete-message-days': delete_message_days
        }
        return self.put(url, params=params, bucket=_func_())

    def unban(self, user_id, guild_id):
        url = '{0.GUILDS}/{1}/bans/{2}'.format(self, guild_id, user_id)
        return self.delete(url, bucket=_func_())

    def server_voice_state(self, user_id, guild_id, *, mute=None, deafen=None):
        url = '{0.GUILDS}/{1}/members/{2}'.format(self, guild_id, user_id)
        payload = {}
        if mute is not None:
            payload['mute'] = mute

        if deafen is not None:
            payload['deaf'] = deafen

        return self.patch(url, json=payload, bucket='members:' + str(guild_id))

    def edit_profile(self, password, username, avatar, **fields):
        payload = {
            'password': password,
            'username': username,
            'avatar': avatar
        }

        if 'email' in fields:
            payload['email'] = fields['email']

        if 'new_password' in fields:
            payload['new_password'] = fields['new_password']

        return self.patch(self.ME, json=payload, bucket=_func_())

    def change_my_nickname(self, guild_id, nickname):
        url = '{0.GUILDS}/{1}/members/@me/nick'.format(self, guild_id)
        payload = {
            'nick': nickname
        }
        bucket = '{}:{}'.format(_func_(), guild_id)
        return self.patch(url, json=payload, bucket=bucket)

    def change_nickname(self, guild_id, user_id, nickname):
        url = '{0.GUILDS}/{1}/members/{2}'.format(self, guild_id, user_id)
        payload = {
            'nick': nickname
        }
        bucket = 'members:{}'.format(guild_id)
        return self.patch(url, json=payload, bucket=bucket)

    # Channel management

    def edit_channel(self, channel_id, **options):
        url = '{0.CHANNELS}/{1}'.format(self, channel_id)

        valid_keys = ('name', 'topic', 'bitrate', 'user_limit', 'position')
        payload = {
            k: v for k, v in options.items() if k in valid_keys
        }

        return self.patch(url, json=payload, bucket=_func_())

    def create_channel(self, guild_id, name, channe_type, permission_overwrites=None):
        url = '{0.GUILDS}/{1}/channels'.format(self, guild_id)
        payload = {
            'name': name,
            'type': channe_type
        }

        if permission_overwrites is not None:
            payload['permission_overwrites'] = permission_overwrites

        return self.post(url, json=payload, bucket=_func_())

    def delete_channel(self, channel_id):
        url = '{0.CHANNELS}/{1}'.format(self, channel_id)
        return self.delete(url, bucket=_func_())

    # Server management

    def leave_server(self, guild_id):
        url = '{0.USERS}/@me/guilds/{1}'.format(self, guild_id)
        return self.delete(url, bucket=_func_())

    def delete_server(self, guild_id):
        url = '{0.GUILDS}/{1}'.format(self, guild_id)
        return self.delete(url, bucket=_func_())

    def create_server(self, name, region, icon):
        payload = {
            'name': name,
            'icon': icon,
            'region': region
        }

        return self.post(self.GUILDS, json=payload, bucket=_func_())

    def edit_server(self, guild_id, **fields):
        valid_keys = ('name', 'region', 'icon', 'afk_timeout', 'owner_id',
                      'afk_channel_id', 'splash', 'verification_level')

        payload = {
            k: v for k, v in fields.items() if k in valid_keys
        }

        url = '{0.GUILDS}/{1}'.format(self, guild_id)
        return self.patch(url, json=payload, bucket=_func_())

    def get_bans(self, guild_id):
        url = '{0.GUILDS}/{1}/bans'.format(self, guild_id)
        return self.get(url, bucket=_func_())

    def prune_members(self, guild_id, days):
        url = '{0.GUILDS}/{1}/prune'.format(self, guild_id)
        params = {
            'days': days
        }
        return self.post(url, params=params, bucket=_func_())

    def estimate_pruned_members(self, guild_id, days):
        url = '{0.GUILDS}/{1}/prune'.format(self, guild_id)
        params = {
            'days': days
        }
        return self.get(url, params=params, bucket=_func_())

    def create_custom_emoji(self, guild_id, name, image):
        payload = {
            'name': name,
            'image': image
        }

        bucket = '%s:%s' % (_func_(), guild_id)
        return self.post('{0.GUILDS}/{1}/emojis'.format(self, guild_id), json=payload, bucket=bucket)

    def delete_custom_emoji(self, guild_id, emoji_id):
        url = '{0.GUILDS}/{1}/emojis/{2}'.format(self, guild_id, emoji_id)
        bucket = '%s:%s' % (_func_(), guild_id)
        return self.delete(url, bucket=bucket)

    def edit_custom_emoji(self, guild_id, emoji_id, *, name):
        payload = {
            'name': name
        }
        url = '{0.GUILDS}/{1}/emojis/{2}'.format(self, guild_id, emoji_id)
        bucket = '%s:%s' % (_func_(), guild_id)
        return self.patch(url, bucket=bucket, json=payload)

    # Invite management

    def create_invite(self, channel_id, **options):
        url = '{0.CHANNELS}/{1}/invites'.format(self, channel_id)
        payload = {
            'max_age': options.get('max_age', 0),
            'max_uses': options.get('max_uses', 0),
            'temporary': options.get('temporary', False),
            'xkcdpass': options.get('xkcd', False)
        }

        return self.post(url, json=payload, bucket=_func_())

    def get_invite(self, invite_id):
        url = '{0.API_BASE}/invite/{1}'.format(self, invite_id)
        return self.get(url, bucket=_func_())

    def invites_from(self, guild_id):
        url = '{0.GUILDS}/{1}/invites'.format(self, guild_id)
        return self.get(url, bucket=_func_())

    def accept_invite(self, invite_id):
        url = '{0.API_BASE}/invite/{1}'.format(self, invite_id)
        return self.post(url, bucket=_func_())

    def delete_invite(self, invite_id):
        url = '{0.API_BASE}/invite/{1}'.format(self, invite_id)
        return self.delete(url, bucket=_func_())

    # Role management

    def edit_role(self, guild_id, role_id, **fields):
        url = '{0.GUILDS}/{1}/roles/{2}'.format(self, guild_id, role_id)
        valid_keys = ('name', 'permissions', 'color', 'hoist', 'mentionable')
        payload = {
            k: v for k, v in fields.items() if k in valid_keys
        }
        return self.patch(url, json=payload, bucket='roles:' + str(guild_id))

    def delete_role(self, guild_id, role_id):
        url = '{0.GUILDS}/{1}/roles/{2}'.format(self, guild_id, role_id)
        return self.delete(url, bucket=_func_())

    def replace_roles(self, user_id, guild_id, role_ids):
        url = '{0.GUILDS}/{1}/members/{2}'.format(self, guild_id, user_id)
        payload = {
            'roles': role_ids
        }
        return self.patch(url, json=payload, bucket='members:' + str(guild_id))

    def create_role(self, guild_id):
        url = '{0.GUILDS}/{1}/roles'.format(self, guild_id)
        return self.post(url, bucket=_func_())

    def add_role(self, guild_id, member_id, role_id):
        url = '{0.GUILDS}/{1}/members/{2}/{3}'.format(self, guild_id, member_id, role_id)
        return self.put(url, bucket='%s:%s' % (_func_(), guild_id))

    def remove_role(self, guild_id, member_id, role_id):
        url = '{0.GUILDS}/{1}/members/{2}/{3}'.format(self, guild_id, member_id, role_id)
        return self.delete(url, bucket='%s:%s' % (_func_(), guild_id))

    def edit_channel_permissions(self, channel_id, target, allow, deny, type):
        url = '{0.CHANNELS}/{1}/permissions/{2}'.format(self, channel_id, target)
        payload = {
            'id': target,
            'allow': allow,
            'deny': deny,
            'type': type
        }
        return self.put(url, json=payload, bucket=_func_())

    def delete_channel_permissions(self, channel_id, target):
        url = '{0.CHANNELS}/{1}/permissions/{2}'.format(self, channel_id, target)
        return self.delete(url, bucket=_func_())

    # Voice management

    def move_member(self, user_id, guild_id, channel_id):
        url = '{0.GUILDS}/{1}/members/{2}'.format(self, guild_id, user_id)
        payload = {
            'channel_id': channel_id
        }
        return self.patch(url, json=payload, bucket='members:' + str(guild_id))

    # Misc

    def application_info(self):
        url = '{0.APPLICATIONS}/@me'.format(self)
        return self.get(url, bucket=_func_())

    @asyncio.coroutine
    def get_gateway(self):
        try:
            data = yield from self.get(self.GATEWAY, bucket=_func_())
        except HTTPException as e:
            raise GatewayNotFound() from e
        return data.get('url') + '?encoding=json&v=6'

    def get_user_info(self, user_id):
        return self.get('{0.USERS}/{1}'.format(self, user_id), bucket=_func_())



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from . import utils
from .user import User
from .reaction import Reaction
from .object import Object
from .calls import CallMessage
import re
from .enums import MessageType, try_enum

class Message:
    """Represents a message from Discord.

    There should be no need to create one of these manually.

    Attributes
    -----------
    edited_timestamp : Optional[datetime.datetime]
        A naive UTC datetime object containing the edited time of the message.
    timestamp : datetime.datetime
        A naive UTC datetime object containing the time the message was created.
    tts : bool
        Specifies if the message was done with text-to-speech.
    type: :class:`MessageType`
        The type of message. In most cases this should not be checked, but it is helpful
        in cases where it might be a system message for :attr:`system_content`.
    author
        A :class:`Member` that sent the message. If :attr:`channel` is a
        private channel, then it is a :class:`User` instead.
    content : str
        The actual contents of the message.
    nonce
        The value used by the discord server and the client to verify that the message is successfully sent.
        This is typically non-important.
    embeds : list
        A list of embedded objects. The elements are objects that meet oEmbed's specification_.

        .. _specification: http://oembed.com/
    channel
        The :class:`Channel` that the message was sent from.
        Could be a :class:`PrivateChannel` if it's a private message.
        In :issue:`very rare cases <21>` this could be a :class:`Object` instead.

        For the sake of convenience, this :class:`Object` instance has an attribute ``is_private`` set to ``True``.
    server : Optional[:class:`Server`]
        The server that the message belongs to. If not applicable (i.e. a PM) then it's None instead.
    call: Optional[:class:`CallMessage`]
        The call that the message refers to. This is only applicable to messages of type
        :attr:`MessageType.call`.
    mention_everyone : bool
        Specifies if the message mentions everyone.

        .. note::

            This does not check if the ``@everyone`` text is in the message itself.
            Rather this boolean indicates if the ``@everyone`` text is in the message
            **and** it did end up mentioning everyone.

    mentions: list
        A list of :class:`Member` that were mentioned. If the message is in a private message
        then the list will be of :class:`User` instead. For messages that are not of type
        :attr:`MessageType.default`\, this array can be used to aid in system messages.
        For more information, see :attr:`system_content`.

        .. warning::

            The order of the mentions list is not in any particular order so you should
            not rely on it. This is a discord limitation, not one with the library.

    channel_mentions : list
        A list of :class:`Channel` that were mentioned. If the message is in a private message
        then the list is always empty.
    role_mentions : list
        A list of :class:`Role` that were mentioned. If the message is in a private message
        then the list is always empty.
    id : str
        The message ID.
    attachments : list
        A list of attachments given to a message.
    pinned: bool
        Specifies if the message is currently pinned.
    reactions : List[:class:`Reaction`]
        Reactions to a message. Reactions can be either custom emoji or standard unicode emoji.
    """

    __slots__ = [ 'edited_timestamp', 'timestamp', 'tts', 'content', 'channel',
                  'mention_everyone', 'embeds', 'id', 'mentions', 'author',
                  'channel_mentions', 'server', '_raw_mentions', 'attachments',
                  '_clean_content', '_raw_channel_mentions', 'nonce', 'pinned',
                  'role_mentions', '_raw_role_mentions', 'type', 'call',
                  '_system_content', 'reactions' ]

    def __init__(self, **kwargs):
        self.reactions = kwargs.pop('reactions')
        for reaction in self.reactions:
            reaction.message = self
        self._update(**kwargs)

    def _update(self, **data):
        # at the moment, the timestamps seem to be naive so they have no time zone and operate on UTC time.
        # we can use this to our advantage to use strptime instead of a complicated parsing routine.
        # example timestamp: 2015-08-21T12:03:45.782000+00:00
        # sometimes the .%f modifier is missing
        self.edited_timestamp = utils.parse_time(data.get('edited_timestamp'))
        self.timestamp = utils.parse_time(data.get('timestamp'))
        self.tts = data.get('tts', False)
        self.pinned = data.get('pinned', False)
        self.content = data.get('content')
        self.mention_everyone = data.get('mention_everyone')
        self.embeds = data.get('embeds')
        self.id = data.get('id')
        self.channel = data.get('channel')
        self.author = User(**data.get('author', {}))
        self.nonce = data.get('nonce')
        self.attachments = data.get('attachments')
        self.type = try_enum(MessageType, data.get('type'))
        self._handle_upgrades(data.get('channel_id'))
        self._handle_mentions(data.get('mentions', []), data.get('mention_roles', []))
        self._handle_call(data.get('call'))

        # clear the cached properties
        cached = filter(lambda attr: attr[0] == '_', self.__slots__)
        for attr in cached:
            try:
                delattr(self, attr)
            except AttributeError:
                pass

    def _handle_mentions(self, mentions, role_mentions):
        self.mentions = []
        self.channel_mentions = []
        self.role_mentions = []
        if getattr(self.channel, 'is_private', True):
            self.mentions = [User(**m) for m in mentions]
            return

        if self.server is not None:
            for mention in mentions:
                id_search = mention.get('id')
                member = self.server.get_member(id_search)
                if member is not None:
                    self.mentions.append(member)

            it = filter(None, map(lambda m: self.server.get_channel(m), self.raw_channel_mentions))
            self.channel_mentions = utils._unique(it)

            for role_id in role_mentions:
                role = utils.get(self.server.roles, id=role_id)
                if role is not None:
                    self.role_mentions.append(role)

    def _handle_call(self, call):
        if call is None or self.type is not MessageType.call:
            self.call = None
            return

        # we get the participant source from the mentions array or
        # the author

        participants = []
        for uid in call.get('participants', []):
            if uid == self.author.id:
                participants.append(self.author)
            else:
                user = utils.find(lambda u: u.id == uid, self.mentions)
                if user is not None:
                    participants.append(user)

        call['participants'] = participants
        self.call = CallMessage(message=self, **call)

    @utils.cached_slot_property('_raw_mentions')
    def raw_mentions(self):
        """A property that returns an array of user IDs matched with
        the syntax of <@user_id> in the message content.

        This allows you receive the user IDs of mentioned users
        even in a private message context.
        """
        return re.findall(r'<@!?([0-9]+)>', self.content)

    @utils.cached_slot_property('_raw_channel_mentions')
    def raw_channel_mentions(self):
        """A property that returns an array of channel IDs matched with
        the syntax of <#channel_id> in the message content.
        """
        return re.findall(r'<#([0-9]+)>', self.content)

    @utils.cached_slot_property('_raw_role_mentions')
    def raw_role_mentions(self):
        """A property that returns an array of role IDs matched with
        the syntax of <@&role_id> in the message content.
        """
        return re.findall(r'<@&([0-9]+)>', self.content)

    @utils.cached_slot_property('_clean_content')
    def clean_content(self):
        """A property that returns the content in a "cleaned up"
        manner. This basically means that mentions are transformed
        into the way the client shows it. e.g. ``<#id>`` will transform
        into ``#name``.

        This will also transform @everyone and @here mentions into
        non-mentions.
        """

        transformations = {
            re.escape('<#{0.id}>'.format(channel)): '#' + channel.name
            for channel in self.channel_mentions
        }

        mention_transforms = {
            re.escape('<@{0.id}>'.format(member)): '@' + member.display_name
            for member in self.mentions
        }

        # add the <@!user_id> cases as well..
        second_mention_transforms = {
            re.escape('<@!{0.id}>'.format(member)): '@' + member.display_name
            for member in self.mentions
        }

        transformations.update(mention_transforms)
        transformations.update(second_mention_transforms)

        if self.server is not None:
            role_transforms = {
                re.escape('<@&{0.id}>'.format(role)): '@' + role.name
                for role in self.role_mentions
            }
            transformations.update(role_transforms)

        def repl(obj):
            return transformations.get(re.escape(obj.group(0)), '')

        pattern = re.compile('|'.join(transformations.keys()))
        result = pattern.sub(repl, self.content)

        transformations = {
            '@everyone': '@\u200beveryone',
            '@here': '@\u200bhere'
        }

        def repl2(obj):
            return transformations.get(obj.group(0), '')

        pattern = re.compile('|'.join(transformations.keys()))
        return pattern.sub(repl2, result)

    def _handle_upgrades(self, channel_id):
        self.server = None
        if isinstance(self.channel, Object):
            return

        if self.channel is None:
            if channel_id is not None:
                self.channel = Object(id=channel_id)
                self.channel.is_private = True
            return

        if not self.channel.is_private:
            self.server = self.channel.server
            found = self.server.get_member(self.author.id)
            if found is not None:
                self.author = found

    @utils.cached_slot_property('_system_content')
    def system_content(self):
        """A property that returns the content that is rendered
        regardless of the :attr:`Message.type`.

        In the case of :attr:`MessageType.default`\, this just returns the
        regular :attr:`Message.content`. Otherwise this returns an English
        message denoting the contents of the system message.
        """

        if self.type is MessageType.default:
            return self.content

        if self.type is MessageType.pins_add:
            return '{0.name} pinned a message to this channel.'.format(self.author)

        if self.type is MessageType.recipient_add:
            return '{0.name} added {1.name} to the group.'.format(self.author, self.mentions[0])

        if self.type is MessageType.recipient_remove:
            return '{0.name} removed {1.name} from the group.'.format(self.author, self.mentions[0])

        if self.type is MessageType.channel_name_change:
            return '{0.author.name} changed the channel name: {0.content}'.format(self)

        if self.type is MessageType.channel_icon_change:
            return '{0.author.name} changed the channel icon.'.format(self)

        if self.type is MessageType.call:
            # we're at the call message type now, which is a bit more complicated.
            # we can make the assumption that Message.channel is a PrivateChannel
            # with the type ChannelType.group or ChannelType.private
            call_ended = self.call.ended_timestamp is not None

            if self.channel.me in self.call.participants:
                return '{0.author.name} started a call.'.format(self)
            elif call_ended:
                return 'You missed a call from {0.author.name}'.format(self)
            else:
                return '{0.author.name} started a call \N{EM DASH} Join the call.'.format(self)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .permissions import Permissions
from .colour import Colour
from .mixins import Hashable
from .utils import snowflake_time

class Role(Hashable):
    """Represents a Discord role in a :class:`Server`.

    Supported Operations:

    +-----------+------------------------------------------------------------------+
    | Operation |                           Description                            |
    +===========+==================================================================+
    | x == y    | Checks if two roles are equal.                                   |
    +-----------+------------------------------------------------------------------+
    | x != y    | Checks if two roles are not equal.                               |
    +-----------+------------------------------------------------------------------+
    | x > y     | Checks if a role is higher than another in the hierarchy.        |
    +-----------+------------------------------------------------------------------+
    | x < y     | Checks if a role is lower than another in the hierarchy.         |
    +-----------+------------------------------------------------------------------+
    | x >= y    | Checks if a role is higher or equal to another in the hierarchy. |
    +-----------+------------------------------------------------------------------+
    | x <= y    | Checks if a role is lower or equal to another in the hierarchy.  |
    +-----------+------------------------------------------------------------------+
    | hash(x)   | Return the role's hash.                                          |
    +-----------+------------------------------------------------------------------+
    | str(x)    | Returns the role's name.                                         |
    +-----------+------------------------------------------------------------------+

    Attributes
    ----------
    id : str
        The ID for the role.
    name : str
        The name of the role.
    permissions : :class:`Permissions`
        Represents the role's permissions.
    server : :class:`Server`
        The server the role belongs to.
    colour : :class:`Colour`
        Represents the role colour. An alias exists under ``color``.
    hoist : bool
         Indicates if the role will be displayed separately from other members.
    position : int
        The position of the role. This number is usually positive. The bottom
        role has a position of 0.
    managed : bool
        Indicates if the role is managed by the server through some form of
        integrations such as Twitch.
    mentionable : bool
        Indicates if the role can be mentioned by users.
    """

    __slots__ = ['id', 'name', 'permissions', 'color', 'colour', 'position',
                 'managed', 'mentionable', 'hoist', 'server' ]

    def __init__(self, **kwargs):
        self.server = kwargs.pop('server')
        self._update(**kwargs)

    def __str__(self):
        return self.name

    def __lt__(self, other):
        if not isinstance(other, Role) or  not isinstance(self, Role):
            return NotImplemented

        if self.server != other.server:
            raise RuntimeError('cannot compare roles from two different servers.')

        if self.position < other.position:
            return True

        if self.position == other.position:
            return self.id > other.id

        return False

    def __le__(self, other):
        r = Role.__lt__(other, self)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def __gt__(self, other):
        return Role.__lt__(other, self)

    def __ge__(self, other):
        r = Role.__lt__(self, other)
        if r is NotImplemented:
            return NotImplemented
        return not r

    def _update(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.permissions = Permissions(kwargs.get('permissions', 0))
        self.position = kwargs.get('position', 0)
        self.colour = Colour(kwargs.get('color', 0))
        self.hoist = kwargs.get('hoist', False)
        self.managed = kwargs.get('managed', False)
        self.mentionable = kwargs.get('mentionable', False)
        self.color = self.colour

    @property
    def is_everyone(self):
        """Checks if the role is the @everyone role."""
        return self.server.id == self.id

    @property
    def created_at(self):
        """Returns the role's creation time in UTC."""
        return snowflake_time(self.id)

    @property
    def mention(self):
        """Returns a string that allows you to mention a role."""
        return '<@&{}>'.format(self.id)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .user import User
from .game import Game
from .permissions import Permissions
from . import utils
from .enums import Status, ChannelType
from .colour import Colour
import copy

class VoiceState:
    """Represents a Discord user's voice state.

    Attributes
    ------------
    deaf: bool
        Indicates if the user is currently deafened by the server.
    mute: bool
        Indicates if the user is currently muted by the server.
    self_mute: bool
        Indicates if the user is currently muted by their own accord.
    self_deaf: bool
        Indicates if the user is currently deafened by their own accord.
    is_afk: bool
        Indicates if the user is currently in the AFK channel in the server.
    voice_channel: Optional[Union[:class:`Channel`, :class:`PrivateChannel`]]
        The voice channel that the user is currently connected to. None if the user
        is not currently in a voice channel.
    """

    __slots__ = [ 'session_id', 'deaf', 'mute', 'self_mute',
                  'self_deaf', 'is_afk', 'voice_channel' ]

    def __init__(self, **kwargs):
        self.session_id = kwargs.get('session_id')
        self._update_voice_state(**kwargs)

    def _update_voice_state(self, **kwargs):
        self.self_mute = kwargs.get('self_mute', False)
        self.self_deaf = kwargs.get('self_deaf', False)
        self.is_afk = kwargs.get('suppress', False)
        self.mute = kwargs.get('mute', False)
        self.deaf = kwargs.get('deaf', False)
        self.voice_channel = kwargs.get('voice_channel')

def flatten_voice_states(cls):
    for attr in VoiceState.__slots__:
        def getter(self, x=attr):
            return getattr(self.voice, x)
        setattr(cls, attr, property(getter))
    return cls

@flatten_voice_states
class Member(User):
    """Represents a Discord member to a :class:`Server`.

    This is a subclass of :class:`User` that extends more functionality
    that server members have such as roles and permissions.

    Attributes
    ----------
    voice: :class:`VoiceState`
        The member's voice state. Properties are defined to mirror access of the attributes.
        e.g. ``Member.is_afk`` is equivalent to `Member.voice.is_afk``.
    roles
        A list of :class:`Role` that the member belongs to. Note that the first element of this
        list is always the default '@everyone' role.
    joined_at : `datetime.datetime`
        A datetime object that specifies the date and time in UTC that the member joined the server for
        the first time.
    status : :class:`Status`
        The member's status. There is a chance that the status will be a ``str``
        if it is a value that is not recognised by the enumerator.
    game : :class:`Game`
        The game that the user is currently playing. Could be None if no game is being played.
    server : :class:`Server`
        The server that the member belongs to.
    nick : Optional[str]
        The server specific nickname of the user.
    """

    __slots__ = [ 'roles', 'joined_at', 'status', 'game', 'server', 'nick', 'voice' ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs.get('user'))
        self.voice = VoiceState(**kwargs)
        self.joined_at = utils.parse_time(kwargs.get('joined_at'))
        self.roles = kwargs.get('roles', [])
        self.status = Status.offline
        game = kwargs.get('game', {})
        self.game = Game(**game) if game else None
        self.server = kwargs.get('server', None)
        self.nick = kwargs.get('nick', None)

    def _update_voice_state(self, **kwargs):
        self.voice.self_mute = kwargs.get('self_mute', False)
        self.voice.self_deaf = kwargs.get('self_deaf', False)
        self.voice.is_afk = kwargs.get('suppress', False)
        self.voice.mute = kwargs.get('mute', False)
        self.voice.deaf = kwargs.get('deaf', False)
        old_channel = getattr(self, 'voice_channel', None)
        vc = kwargs.get('voice_channel')

        if old_channel is None and vc is not None:
            # we joined a channel
            vc.voice_members.append(self)
        elif old_channel is not None:
            try:
                # we either left a channel or we switched channels
                old_channel.voice_members.remove(self)
            except ValueError:
                pass
            finally:
                # we switched channels
                if vc is not None:
                    vc.voice_members.append(self)

        self.voice.voice_channel = vc

    def _copy(self):
        ret = copy.copy(self)
        ret.voice = copy.copy(self.voice)
        return ret

    @property
    def colour(self):
        """A property that returns a :class:`Colour` denoting the rendered colour
        for the member. If the default colour is the one rendered then an instance
        of :meth:`Colour.default` is returned.

        There is an alias for this under ``color``.
        """

        default_colour = Colour.default()
        # highest order of the colour is the one that gets rendered.
        # if the highest is the default colour then the next one with a colour
        # is chosen instead
        if self.roles:
            roles = sorted(self.roles, key=lambda r: r.position, reverse=True)
            for role in roles:
                if role.colour == default_colour:
                    continue
                else:
                    return role.colour

        return default_colour

    color = colour

    @property
    def mention(self):
        if self.nick:
            return '<@!{}>'.format(self.id)
        return '<@{}>'.format(self.id)

    def mentioned_in(self, message):
        mentioned = super().mentioned_in(message)
        if mentioned:
            return True

        for role in message.role_mentions:
            has_role = utils.get(self.roles, id=role.id) is not None
            if has_role:
                return True

        return False

    @property
    def top_role(self):
        """Returns the member's highest role.

        This is useful for figuring where a member stands in the role
        hierarchy chain.
        """

        if self.roles:
            roles = sorted(self.roles, reverse=True)
            return roles[0]
        return None

    @property
    def server_permissions(self):
        """Returns the member's server permissions.

        This only takes into consideration the server permissions
        and not most of the implied permissions or any of the
        channel permission overwrites. For 100% accurate permission
        calculation, please use either :meth:`permissions_in` or
        :meth:`Channel.permissions_for`.

        This does take into consideration server ownership and the
        administrator implication.
        """

        if self.server.owner == self:
            return Permissions.all()

        base = Permissions.none()
        for r in self.roles:
            base.value |= r.permissions.value

        if base.administrator:
            return Permissions.all()

        return base



# -*- coding: utf-8 -*-
"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import concurrent.futures
import asyncio

try:
    create_task = asyncio.ensure_future
except AttributeError:
    create_task = asyncio.async

try:
    run_coroutine_threadsafe = asyncio.run_coroutine_threadsafe
except AttributeError:
    # the following code is slightly modified from the
    # official asyncio repository that could be found here:
    # https://github.com/python/asyncio/blob/master/asyncio/futures.py
    # with a commit hash of 5c7efbcdfbe6a5c25b4cd5df22d9a15ab4062c8e
    # this portion is licensed under Apache license 2.0

    def _set_concurrent_future_state(concurrent, source):
        """Copy state from a future to a concurrent.futures.Future."""
        assert source.done()
        if source.cancelled():
            concurrent.cancel()
        if not concurrent.set_running_or_notify_cancel():
            return
        exception = source.exception()
        if exception is not None:
            concurrent.set_exception(exception)
        else:
            result = source.result()
            concurrent.set_result(result)


    def _copy_future_state(source, dest):
        """Internal helper to copy state from another Future.
        The other Future may be a concurrent.futures.Future.
        """
        assert source.done()
        if dest.cancelled():
            return
        assert not dest.done()
        if source.cancelled():
            dest.cancel()
        else:
            exception = source.exception()
            if exception is not None:
                dest.set_exception(exception)
            else:
                result = source.result()
                dest.set_result(result)

    def _chain_future(source, destination):
        """Chain two futures so that when one completes, so does the other.
        The result (or exception) of source will be copied to destination.
        If destination is cancelled, source gets cancelled too.
        Compatible with both asyncio.Future and concurrent.futures.Future.
        """
        if not isinstance(source, (asyncio.Future, concurrent.futures.Future)):
            raise TypeError('A future is required for source argument')

        if not isinstance(destination, (asyncio.Future, concurrent.futures.Future)):
            raise TypeError('A future is required for destination argument')

        source_loop = source._loop if isinstance(source, asyncio.Future) else None
        dest_loop = destination._loop if isinstance(destination, asyncio.Future) else None

        def _set_state(future, other):
            if isinstance(future, asyncio.Future):
                _copy_future_state(other, future)
            else:
                _set_concurrent_future_state(future, other)

        def _call_check_cancel(destination):
            if destination.cancelled():
                if source_loop is None or source_loop is dest_loop:
                    source.cancel()
                else:
                    source_loop.call_soon_threadsafe(source.cancel)

        def _call_set_state(source):
            if dest_loop is None or dest_loop is source_loop:
                _set_state(destination, source)
            else:
                dest_loop.call_soon_threadsafe(_set_state, destination, source)

        destination.add_done_callback(_call_check_cancel)
        source.add_done_callback(_call_set_state)

    def run_coroutine_threadsafe(coro, loop):
        """Submit a coroutine object to a given event loop.

        Return a concurrent.futures.Future to access the result.
        """
        if not asyncio.iscoroutine(coro):
            raise TypeError('A coroutine object is required')

        future = concurrent.futures.Future()

        def callback():
            try:
                _chain_future(create_task(coro, loop=loop), future)
            except Exception as exc:
                if future.set_running_or_notify_cancel():
                    future.set_exception(exc)
                raise
        loop.call_soon_threadsafe(callback)
        return future



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .server import Server
from .user import User
from .game import Game
from .emoji import Emoji
from .reaction import Reaction
from .message import Message
from .channel import Channel, PrivateChannel
from .member import Member
from .role import Role
from . import utils, compat
from .enums import Status, ChannelType, try_enum
from .calls import GroupCall

from collections import deque, namedtuple
import copy, enum, math
import datetime
import asyncio
import logging

class ListenerType(enum.Enum):
    chunk = 0

Listener = namedtuple('Listener', ('type', 'future', 'predicate'))
log = logging.getLogger(__name__)
ReadyState = namedtuple('ReadyState', ('launch', 'servers'))

class ConnectionState:
    def __init__(self, dispatch, chunker, syncer, max_messages, *, loop):
        self.loop = loop
        self.max_messages = max_messages
        self.dispatch = dispatch
        self.chunker = chunker
        self.syncer = syncer
        self.is_bot = None
        self._listeners = []
        self.clear()

    def clear(self):
        self.user = None
        self.sequence = None
        self.session_id = None
        self._calls = {}
        self._servers = {}
        self._voice_clients = {}
        self._private_channels = {}
        # extra dict to look up private channels by user id
        self._private_channels_by_user = {}
        self.messages = deque(maxlen=self.max_messages)

    def process_listeners(self, listener_type, argument, result):
        removed = []
        for i, listener in enumerate(self._listeners):
            if listener.type != listener_type:
                continue

            future = listener.future
            if future.cancelled():
                removed.append(i)
                continue

            try:
                passed = listener.predicate(argument)
            except Exception as e:
                future.set_exception(e)
                removed.append(i)
            else:
                if passed:
                    future.set_result(result)
                    removed.append(i)
                    if listener.type == ListenerType.chunk:
                        break

        for index in reversed(removed):
            del self._listeners[index]

    @property
    def voice_clients(self):
        return self._voice_clients.values()

    def _get_voice_client(self, guild_id):
        return self._voice_clients.get(guild_id)

    def _add_voice_client(self, guild_id, voice):
        self._voice_clients[guild_id] = voice

    def _remove_voice_client(self, guild_id):
        self._voice_clients.pop(guild_id, None)

    def _update_references(self, ws):
        for vc in self.voice_clients:
            vc.main_ws = ws

    @property
    def servers(self):
        return self._servers.values()

    def _get_server(self, server_id):
        return self._servers.get(server_id)

    def _add_server(self, server):
        self._servers[server.id] = server

    def _remove_server(self, server):
        self._servers.pop(server.id, None)

    @property
    def private_channels(self):
        return self._private_channels.values()

    def _get_private_channel(self, channel_id):
        return self._private_channels.get(channel_id)

    def _get_private_channel_by_user(self, user_id):
        return self._private_channels_by_user.get(user_id)

    def _add_private_channel(self, channel):
        self._private_channels[channel.id] = channel
        if channel.type is ChannelType.private:
            self._private_channels_by_user[channel.user.id] = channel

    def _remove_private_channel(self, channel):
        self._private_channels.pop(channel.id, None)
        if channel.type is ChannelType.private:
            self._private_channels_by_user.pop(channel.user.id, None)

    def _get_message(self, msg_id):
        return utils.find(lambda m: m.id == msg_id, self.messages)

    def _add_server_from_data(self, guild):
        server = Server(**guild)
        Server.me = property(lambda s: s.get_member(self.user.id))
        Server.voice_client = property(lambda s: self._get_voice_client(s.id))
        self._add_server(server)
        return server

    def chunks_needed(self, server):
        for chunk in range(math.ceil(server._member_count / 1000)):
            yield self.receive_chunk(server.id)

    @asyncio.coroutine
    def _delay_ready(self):
        launch = self._ready_state.launch
        while not launch.is_set():
            # this snippet of code is basically waiting 2 seconds
            # until the last GUILD_CREATE was sent
            launch.set()
            yield from asyncio.sleep(2)

        servers = self._ready_state.servers

        # get all the chunks
        chunks = []
        for server in servers:
            chunks.extend(self.chunks_needed(server))

        # we only want to request ~75 guilds per chunk request.
        splits = [servers[i:i + 75] for i in range(0, len(servers), 75)]
        for split in splits:
            yield from self.chunker(split)

        # wait for the chunks
        if chunks:
            yield from asyncio.wait(chunks)

        # remove the state
        del self._ready_state

        # call GUILD_SYNC after we're done chunking
        if not self.is_bot:
            compat.create_task(self.syncer([s.id for s in self.servers]), loop=self.loop)

        # dispatch the event
        self.dispatch('ready')

    def parse_ready(self, data):
        self._ready_state = ReadyState(launch=asyncio.Event(), servers=[])
        self.user = User(**data['user'])
        guilds = data.get('guilds')

        servers = self._ready_state.servers
        for guild in guilds:
            server = self._add_server_from_data(guild)
            if server.large or not self.is_bot:
                servers.append(server)

        for pm in data.get('private_channels'):
            self._add_private_channel(PrivateChannel(self.user, **pm))

        compat.create_task(self._delay_ready(), loop=self.loop)

    def parse_resumed(self, data):
        self.dispatch('resumed')

    def parse_message_create(self, data):
        channel = self.get_channel(data.get('channel_id'))
        message = self._create_message(channel=channel, **data)
        self.dispatch('message', message)
        self.messages.append(message)

    def parse_message_delete(self, data):
        message_id = data.get('id')
        found = self._get_message(message_id)
        if found is not None:
            self.dispatch('message_delete', found)
            self.messages.remove(found)

    def parse_message_delete_bulk(self, data):
        message_ids = set(data.get('ids', []))
        to_be_deleted = list(filter(lambda m: m.id in message_ids, self.messages))
        for msg in to_be_deleted:
            self.dispatch('message_delete', msg)
            self.messages.remove(msg)

    def parse_message_update(self, data):
        message = self._get_message(data.get('id'))
        if message is not None:
            older_message = copy.copy(message)
            if 'call' in data:
                # call state message edit
                message._handle_call(data['call'])
            elif 'content' not in data:
                # embed only edit
                message.embeds = data['embeds']
            else:
                message._update(channel=message.channel, **data)

            self.dispatch('message_edit', older_message, message)

    def parse_message_reaction_add(self, data):
        message = self._get_message(data['message_id'])
        if message is not None:
            emoji = self._get_reaction_emoji(**data.pop('emoji'))
            reaction = utils.get(message.reactions, emoji=emoji)

            is_me = data['user_id'] == self.user.id

            if not reaction:
                reaction = Reaction(
                    message=message, emoji=emoji, me=is_me, **data)
                message.reactions.append(reaction)
            else:
                reaction.count += 1
                if is_me:
                    reaction.me = True

            channel = self.get_channel(data['channel_id'])
            member = self._get_member(channel, data['user_id'])

            self.dispatch('reaction_add', reaction, member)

    def parse_message_reaction_remove_all(self, data):
        message =  self._get_message(data['message_id'])
        if message is not None:
            old_reactions = message.reactions.copy()
            message.reactions.clear()
            self.dispatch('reaction_clear', message, old_reactions)

    def parse_message_reaction_remove(self, data):
        message = self._get_message(data['message_id'])
        if message is not None:
            emoji = self._get_reaction_emoji(**data['emoji'])
            reaction = utils.get(message.reactions, emoji=emoji)

            # if reaction isn't in the list, we crash. This means discord
            # sent bad data, or we stored improperly
            reaction.count -= 1
            if data['user_id'] == self.user.id:
                reaction.me = False
            if reaction.count == 0:
                message.reactions.remove(reaction)

            channel = self.get_channel(data['channel_id'])
            member = self._get_member(channel, data['user_id'])

            self.dispatch('reaction_remove', reaction, member)

    def parse_presence_update(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is None:
            return

        status = data.get('status')
        user = data['user']
        member_id = user['id']
        member = server.get_member(member_id)
        if member is None:
            if 'username' not in user:
                # sometimes we receive 'incomplete' member data post-removal.
                # skip these useless cases.
                return

            member = self._make_member(server, data)
            server._add_member(member)

        old_member = member._copy()
        member.status = data.get('status')
        try:
            member.status = Status(member.status)
        except:
            pass

        game = data.get('game', {})
        member.game = Game(**game) if game else None
        member.name = user.get('username', member.name)
        member.avatar = user.get('avatar', member.avatar)
        member.discriminator = user.get('discriminator', member.discriminator)

        self.dispatch('member_update', old_member, member)

    def parse_user_update(self, data):
        self.user = User(**data)

    def parse_channel_delete(self, data):
        server =  self._get_server(data.get('guild_id'))
        if server is not None:
            channel_id = data.get('id')
            channel = server.get_channel(channel_id)
            if channel is not None:
                server._remove_channel(channel)
                self.dispatch('channel_delete', channel)

    def parse_channel_update(self, data):
        channel_type = try_enum(ChannelType, data.get('type'))
        channel_id = data.get('id')
        if channel_type is ChannelType.group:
            channel = self._get_private_channel(channel_id)
            old_channel = copy.copy(channel)
            channel._update_group(**data)
            self.dispatch('channel_update', old_channel, channel)
            return

        server = self._get_server(data.get('guild_id'))
        if server is not None:
            channel = server.get_channel(channel_id)
            if channel is not None:
                old_channel = copy.copy(channel)
                channel._update(server=server, **data)
                self.dispatch('channel_update', old_channel, channel)

    def parse_channel_create(self, data):
        ch_type = try_enum(ChannelType, data.get('type'))
        channel = None
        if ch_type in (ChannelType.group, ChannelType.private):
            channel = PrivateChannel(self.user, **data)
            self._add_private_channel(channel)
        else:
            server = self._get_server(data.get('guild_id'))
            if server is not None:
                channel = Channel(server=server, **data)
                server._add_channel(channel)

        self.dispatch('channel_create', channel)

    def parse_channel_recipient_add(self, data):
        channel = self._get_private_channel(data.get('channel_id'))
        user = User(**data.get('user', {}))
        channel.recipients.append(user)
        self.dispatch('group_join', channel, user)

    def parse_channel_recipient_remove(self, data):
        channel = self._get_private_channel(data.get('channel_id'))
        user = User(**data.get('user', {}))
        try:
            channel.recipients.remove(user)
        except ValueError:
            pass
        else:
            self.dispatch('group_remove', channel, user)

    def _make_member(self, server, data):
        roles = [server.default_role]
        for roleid in data.get('roles', []):
            role = utils.get(server.roles, id=roleid)
            if role is not None:
                roles.append(role)

        data['roles'] = sorted(roles, key=lambda r: int(r.id))
        return Member(server=server, **data)

    def parse_guild_member_add(self, data):
        server = self._get_server(data.get('guild_id'))
        member = self._make_member(server, data)
        server._add_member(member)
        server._member_count += 1
        self.dispatch('member_join', member)

    def parse_guild_member_remove(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            user_id = data['user']['id']
            member = server.get_member(user_id)
            if member is not None:
                server._remove_member(member)
                server._member_count -= 1

                # remove them from the voice channel member list
                vc = member.voice_channel
                if vc is not None:
                    try:
                        vc.voice_members.remove(member)
                    except:
                        pass

                self.dispatch('member_remove', member)

    def parse_guild_member_update(self, data):
        server = self._get_server(data.get('guild_id'))
        user_id = data['user']['id']
        member = server.get_member(user_id)
        if member is not None:
            user = data['user']
            old_member = member._copy()
            member.name = user['username']
            member.discriminator = user['discriminator']
            member.avatar = user['avatar']
            member.bot = user.get('bot', False)

            # the nickname change is optional,
            # if it isn't in the payload then it didn't change
            if 'nick' in data:
                member.nick = data['nick']

            # update the roles
            member.roles = [server.default_role]
            for role in server.roles:
                if role.id in data['roles']:
                    member.roles.append(role)

            # sort the roles by ID since they can be "randomised"
            member.roles.sort(key=lambda r: int(r.id))
            self.dispatch('member_update', old_member, member)

    def parse_guild_emojis_update(self, data):
        server = self._get_server(data.get('guild_id'))
        before_emojis = server.emojis
        server.emojis = [Emoji(server=server, **e) for e in data.get('emojis', [])]
        self.dispatch('server_emojis_update', before_emojis, server.emojis)

    def _get_create_server(self, data):
        if data.get('unavailable') == False:
            # GUILD_CREATE with unavailable in the response
            # usually means that the server has become available
            # and is therefore in the cache
            server = self._get_server(data.get('id'))
            if server is not None:
                server.unavailable = False
                server._from_data(data)
                return server

        return self._add_server_from_data(data)

    @asyncio.coroutine
    def _chunk_and_dispatch(self, server, unavailable):
        yield from self.chunker(server)
        chunks = list(self.chunks_needed(server))
        if chunks:
            yield from asyncio.wait(chunks)

        if unavailable == False:
            self.dispatch('server_available', server)
        else:
            self.dispatch('server_join', server)

    def parse_guild_create(self, data):
        unavailable = data.get('unavailable')
        if unavailable == True:
            # joined a server with unavailable == True so..
            return

        server = self._get_create_server(data)

        # check if it requires chunking
        if server.large:
            if unavailable == False:
                # check if we're waiting for 'useful' READY
                # and if we are, we don't want to dispatch any
                # event such as server_join or server_available
                # because we're still in the 'READY' phase. Or
                # so we say.
                try:
                    state = self._ready_state
                    state.launch.clear()
                    state.servers.append(server)
                except AttributeError:
                    # the _ready_state attribute is only there during
                    # processing of useful READY.
                    pass
                else:
                    return

            # since we're not waiting for 'useful' READY we'll just
            # do the chunk request here
            compat.create_task(self._chunk_and_dispatch(server, unavailable), loop=self.loop)
            return

        # Dispatch available if newly available
        if unavailable == False:
            self.dispatch('server_available', server)
        else:
            self.dispatch('server_join', server)

    def parse_guild_sync(self, data):
        server = self._get_server(data.get('id'))
        server._sync(data)

    def parse_guild_update(self, data):
        server = self._get_server(data.get('id'))
        if server is not None:
            old_server = copy.copy(server)
            server._from_data(data)
            self.dispatch('server_update', old_server, server)

    def parse_guild_delete(self, data):
        server = self._get_server(data.get('id'))
        if server is None:
            return

        if data.get('unavailable', False) and server is not None:
            # GUILD_DELETE with unavailable being True means that the
            # server that was available is now currently unavailable
            server.unavailable = True
            self.dispatch('server_unavailable', server)
            return

        # do a cleanup of the messages cache
        self.messages = deque((msg for msg in self.messages if msg.server != server), maxlen=self.max_messages)

        self._remove_server(server)
        self.dispatch('server_remove', server)

    def parse_guild_ban_add(self, data):
        # we make the assumption that GUILD_BAN_ADD is done
        # before GUILD_MEMBER_REMOVE is called
        # hence we don't remove it from cache or do anything
        # strange with it, the main purpose of this event
        # is mainly to dispatch to another event worth listening to for logging
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            user_id = data.get('user', {}).get('id')
            member = utils.get(server.members, id=user_id)
            if member is not None:
                self.dispatch('member_ban', member)

    def parse_guild_ban_remove(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            if 'user' in data:
                user = User(**data['user'])
                self.dispatch('member_unban', server, user)

    def parse_guild_role_create(self, data):
        server = self._get_server(data.get('guild_id'))
        role_data = data.get('role', {})
        role = Role(server=server, **role_data)
        server._add_role(role)
        self.dispatch('server_role_create', role)

    def parse_guild_role_delete(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            role_id = data.get('role_id')
            role = utils.find(lambda r: r.id == role_id, server.roles)
            try:
                server._remove_role(role)
            except ValueError:
                return
            else:
                self.dispatch('server_role_delete', role)

    def parse_guild_role_update(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            role_id = data['role']['id']
            role = utils.find(lambda r: r.id == role_id, server.roles)
            if role is not None:
                old_role = copy.copy(role)
                role._update(**data['role'])
                self.dispatch('server_role_update', old_role, role)

    def parse_guild_members_chunk(self, data):
        server = self._get_server(data.get('guild_id'))
        members = data.get('members', [])
        for member in members:
            m = self._make_member(server, member)
            existing = server.get_member(m.id)
            if existing is None or existing.joined_at is None:
                server._add_member(m)

        # if the owner is offline, server.owner is potentially None
        # therefore we should check if this chunk makes it point to a valid
        # member.
        server.owner = server.get_member(server.owner_id)
        log.info('processed a chunk for {} members.'.format(len(members)))
        self.process_listeners(ListenerType.chunk, server, len(members))

    def parse_voice_state_update(self, data):
        server = self._get_server(data.get('guild_id'))
        if server is not None:
            channel = server.get_channel(data.get('channel_id'))
            if data.get('user_id') == self.user.id:
                voice = self._get_voice_client(server.id)
                if voice is not None:
                    voice.channel = channel

            before, after = server._update_voice_state(data)
            if after is not None:
                self.dispatch('voice_state_update', before, after)
        else:
            # in here we're either at private or group calls
            call = self._calls.get(data.get('channel_id'), None)
            if call is not None:
                call._update_voice_state(data)

    def parse_typing_start(self, data):
        channel = self.get_channel(data.get('channel_id'))
        if channel is not None:
            member = None
            user_id = data.get('user_id')
            is_private = getattr(channel, 'is_private', None)
            if is_private == None:
                return

            if is_private:
                member = channel.user
            else:
                member = channel.server.get_member(user_id)

            if member is not None:
                timestamp = datetime.datetime.utcfromtimestamp(data.get('timestamp'))
                self.dispatch('typing', channel, member, timestamp)

    def parse_call_create(self, data):
        message = self._get_message(data.get('message_id'))
        if message is not None:
            call = GroupCall(call=message, **data)
            self._calls[data['channel_id']] = call
            self.dispatch('call', call)

    def parse_call_update(self, data):
        call = self._calls.get(data.get('channel_id'), None)
        if call is not None:
            before = copy.copy(call)
            call._update(**data)
            self.dispatch('call_update', before, call)

    def parse_call_delete(self, data):
        call = self._calls.pop(data.get('channel_id'), None)
        if call is not None:
            self.dispatch('call_remove', call)

    def _get_member(self, channel, id):
        if channel.is_private:
            return utils.get(channel.recipients, id=id)
        else:
            return channel.server.get_member(id)

    def _create_message(self, **message):
        """Helper mostly for injecting reactions."""
        reactions = [
            self._create_reaction(**r) for r in message.pop('reactions', [])
        ]
        return Message(channel=message.pop('channel'),
                       reactions=reactions, **message)

    def _create_reaction(self, **reaction):
        emoji = self._get_reaction_emoji(**reaction.pop('emoji'))
        return Reaction(emoji=emoji, **reaction)

    def _get_reaction_emoji(self, **data):
        id = data['id']

        if not id:
            return data['name']

        for server in self.servers:
            for emoji in server.emojis:
                if emoji.id == id:
                    return emoji
        return Emoji(server=None, **data)

    def get_channel(self, id):
        if id is None:
            return None

        for server in self.servers:
            channel = server.get_channel(id)
            if channel is not None:
                return channel

        pm = self._get_private_channel(id)
        if pm is not None:
            return pm

    def receive_chunk(self, guild_id):
        future = asyncio.Future(loop=self.loop)
        listener = Listener(ListenerType.chunk, future, lambda s: s.id == guild_id)
        self._listeners.append(listener)
        return future



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from enum import Enum

class ChannelType(Enum):
    text    = 0
    private = 1
    voice   = 2
    group   = 3

    def __str__(self):
        return self.name

class MessageType(Enum):
    default             = 0
    recipient_add       = 1
    recipient_remove    = 2
    call                = 3
    channel_name_change = 4
    channel_icon_change = 5
    pins_add            = 6

class ServerRegion(Enum):
    us_west       = 'us-west'
    us_east       = 'us-east'
    us_south      = 'us-south'
    us_central    = 'us-central'
    eu_west       = 'eu-west'
    eu_central    = 'eu-central'
    singapore     = 'singapore'
    london        = 'london'
    sydney        = 'sydney'
    amsterdam     = 'amsterdam'
    frankfurt     = 'frankfurt'
    brazil        = 'brazil'
    vip_us_east   = 'vip-us-east'
    vip_us_west   = 'vip-us-west'
    vip_amsterdam = 'vip-amsterdam'

    def __str__(self):
        return self.value

class VerificationLevel(Enum):
    none       = 0
    low        = 1
    medium     = 2
    high       = 3
    table_flip = 3

    def __str__(self):
        return self.name

class Status(Enum):
    online = 'online'
    offline = 'offline'
    idle = 'idle'
    dnd = 'dnd'
    do_not_disturb = 'dnd'
    invisible = 'invisible'

    def __str__(self):
        return self.value

class DefaultAvatar(Enum):
    blurple = "6debd47ed13483642cf09e832ed0bc1b"
    grey = "322c936a8c8be1b803cd94861bdfa868"
    green = "dd4dbc0016779df1378e7812eabaa04d"
    orange = "0e291f67c9274a1abdddeb3fd919cbaa"
    red = "1cbd08c76f8af6dddce02c5138971129"

    def __new__(cls, url):
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj._value_ = value
        obj.url = url
        return obj

    def __str__(self):
        return self.name

def try_enum(cls, val):
    """A function that tries to turn the value into enum ``cls``.

    If it fails it returns the value instead.
    """
    try:
        return cls(val)
    except ValueError:
        return val



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

class Game:
    """Represents a Discord game.

    Supported Operations:

    +-----------+------------------------------------+
    | Operation |            Description             |
    +===========+====================================+
    | x == y    | Checks if two games are equal.     |
    +-----------+------------------------------------+
    | x != y    | Checks if two games are not equal. |
    +-----------+------------------------------------+
    | hash(x)   | Return the games's hash.           |
    +-----------+------------------------------------+
    | str(x)    | Returns the games's name.          |
    +-----------+------------------------------------+

    Attributes
    -----------
    name : str
        The game's name.
    url : str
        The game's URL. Usually used for twitch streaming.
    type : int
        The type of game being played. 1 indicates "Streaming".
    """

    __slots__ = ['name', 'type', 'url']

    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.url = kwargs.get('url')
        self.type = kwargs.get('type')

    def __str__(self):
        return self.name

    def _iterator(self):
        for attr in self.__slots__:
            value = getattr(self, attr, None)
            if value is not None:
                yield (attr, value)

    def __iter__(self):
        return self._iterator()

    def __eq__(self, other):
        return isinstance(other, Game) and other.name == self.name

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.name)



# -*- coding: utf-8 -*-
"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import enum
import time

__all__ = ['BucketType', 'Cooldown', 'CooldownMapping']

class BucketType(enum.Enum):
    default = 0
    user    = 1
    server  = 2
    channel = 3

class Cooldown:
    __slots__ = ['rate', 'per', 'type', '_window', '_tokens', '_last']

    def __init__(self, rate, per, type):
        self.rate = int(rate)
        self.per = float(per)
        self.type = type
        self._window = 0.0
        self._tokens = self.rate
        self._last = 0.0

        if not isinstance(self.type, BucketType):
            raise TypeError('Cooldown type must be a BucketType')

    def is_rate_limited(self):
        current = time.time()
        self._last = current

        # first token used means that we start a new rate limit window
        if self._tokens == self.rate:
            self._window = current

        # check if our window has passed and we can refresh our tokens
        if current > self._window + self.per:
            self._tokens = self.rate
            self._window = current

        # check if we're rate limited
        if self._tokens == 0:
            return self.per - (current - self._window)

        # we're not so decrement our tokens
        self._tokens -= 1

        # see if we got rate limited due to this token change, and if
        # so update the window to point to our current time frame
        if self._tokens == 0:
            self._window = current

    def reset(self):
        self._tokens = self.rate
        self._last = 0.0

    def copy(self):
        return Cooldown(self.rate, self.per, self.type)

    def __repr__(self):
        return '<Cooldown rate: {0.rate} per: {0.per} window: {0._window} tokens: {0._tokens}>'.format(self)

class CooldownMapping:
    def __init__(self, original):
        self._cache = {}
        self._cooldown = original

    @property
    def valid(self):
        return self._cooldown is not None

    def _bucket_key(self, ctx):
        msg = ctx.message
        bucket_type = self._cooldown.type
        if bucket_type is BucketType.user:
            return msg.author.id
        elif bucket_type is BucketType.server:
            return getattr(msg.server, 'id', msg.author.id)
        elif bucket_type is BucketType.channel:
            return msg.channel.id

    def _verify_cache_integrity(self):
        # we want to delete all cache objects that haven't been used
        # in a cooldown window. e.g. if we have a  command that has a
        # cooldown of 60s and it has not been used in 60s then that key should be deleted
        current = time.time()
        dead_keys = [k for k, v in self._cache.items() if current > v._last + v.per]
        for k in dead_keys:
            del self._cache[k]

    def get_bucket(self, ctx):
        if self._cooldown.type is BucketType.default:
            return self._cooldown

        self._verify_cache_integrity()
        key = self._bucket_key(ctx)
        if key not in self._cache:
            bucket = self._cooldown.copy()
            self._cache[key] = bucket
        else:
            bucket = self._cache[key]

        return bucket



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import discord
import asyncio
import re
import inspect

from .errors import BadArgument, NoPrivateMessage

__all__ = [ 'Converter', 'MemberConverter', 'UserConverter',
            'ChannelConverter', 'InviteConverter', 'RoleConverter',
            'GameConverter', 'ColourConverter' ]

def _get_from_servers(bot, getter, argument):
    result = None
    for server in bot.servers:
        result = getattr(server, getter)(argument)
        if result:
            return result
    return result

class Converter:
    """The base class of custom converters that require the :class:`Context`
    to be passed to be useful.

    This allows you to implement converters that function similar to the
    special cased ``discord`` classes.

    Classes that derive from this should override the :meth:`convert` method
    to do its conversion logic. This method could be a coroutine or a regular
    function.

    Attributes
    -----------
    ctx: :class:`Context`
        The invocation context that the argument is being used in.
    argument: str
        The argument that is being converted.
    """
    def __init__(self, ctx, argument):
        self.ctx = ctx
        self.argument = argument

    def convert(self):
        raise NotImplementedError('Derived classes need to implement this.')

class IDConverter(Converter):
    def __init__(self, ctx, argument):
        super().__init__(ctx, argument)
        self._id_regex = re.compile(r'([0-9]{15,21})$')

    def _get_id_match(self):
        return self._id_regex.match(self.argument)

class MemberConverter(IDConverter):
    def convert(self):
        message = self.ctx.message
        bot = self.ctx.bot
        match = self._get_id_match() or re.match(r'<@!?([0-9]+)>$', self.argument)
        server = message.server
        result = None
        if match is None:
            # not a mention...
            if server:
                result = server.get_member_named(self.argument)
            else:
                result = _get_from_servers(bot, 'get_member_named', self.argument)
        else:
            user_id = match.group(1)
            if server:
                result = server.get_member(user_id)
            else:
                result = _get_from_servers(bot, 'get_member', user_id)

        if result is None:
            raise BadArgument('Member "{}" not found'.format(self.argument))

        return result

UserConverter = MemberConverter

class ChannelConverter(IDConverter):
    def convert(self):
        message = self.ctx.message
        bot = self.ctx.bot

        match = self._get_id_match() or re.match(r'<#([0-9]+)>$', self.argument)
        result = None
        server = message.server
        if match is None:
            # not a mention
            if server:
                result = discord.utils.get(server.channels, name=self.argument)
            else:
                result = discord.utils.get(bot.get_all_channels(), name=self.argument)
        else:
            channel_id = match.group(1)
            if server:
                result = server.get_channel(channel_id)
            else:
                result = _get_from_servers(bot, 'get_channel', channel_id)

        if result is None:
            raise BadArgument('Channel "{}" not found.'.format(self.argument))

        return result

class ColourConverter(Converter):
    def convert(self):
        arg = self.argument.replace('0x', '').lower()

        if arg[0] == '#':
            arg = arg[1:]
        try:
            value = int(arg, base=16)
            return discord.Colour(value=value)
        except ValueError:
            method = getattr(discord.Colour, arg, None)
            if method is None or not inspect.ismethod(method):
                raise BadArgument('Colour "{}" is invalid.'.format(arg))
            return method()

class RoleConverter(IDConverter):
    def convert(self):
        server = self.ctx.message.server
        if not server:
            raise NoPrivateMessage()

        match = self._get_id_match() or re.match(r'<@&([0-9]+)>$', self.argument)
        params = dict(id=match.group(1)) if match else dict(name=self.argument)
        result = discord.utils.get(server.roles, **params)
        if result is None:
            raise BadArgument('Role "{}" not found.'.format(self.argument))
        return result

class GameConverter(Converter):
    def convert(self):
        return discord.Game(name=self.argument)

class InviteConverter(Converter):
    @asyncio.coroutine
    def convert(self):
        try:
            invite = yield from self.ctx.bot.get_invite(self.argument)
            return invite
        except Exception as e:
            raise BadArgument('Invite is invalid or expired') from e

class EmojiConverter(IDConverter):
    @asyncio.coroutine
    def convert(self):
        message = self.ctx.message
        bot = self.ctx.bot

        match = self._get_id_match() or re.match(r'<:[a-zA-Z0-9]+:([0-9]+)>$', self.argument)
        result = None
        server = message.server
        if match is None:
            # Try to get the emoji by name. Try local server first.
            if server:
                result = discord.utils.get(server.emojis, name=self.argument)

            if result is None:
                result = discord.utils.get(bot.get_all_emojis(), name=self.argument)
        else:
            emoji_id = match.group(1)

            # Try to look up emoji by id.
            if server:
                result = discord.utils.get(server.emojis, id=emoji_id)

            if result is None:
                result = discord.utils.get(bot.get_all_emojis(), id=emoji_id)

        if result is None:
            raise BadArgument('Emoji "{}" not found.'.format(self.argument))

        return result



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .utils import snowflake_time
from .enums import DefaultAvatar

class User:
    """Represents a Discord user.

    Supported Operations:

    +-----------+---------------------------------------------+
    | Operation |                 Description                 |
    +===========+=============================================+
    | x == y    | Checks if two users are equal.              |
    +-----------+---------------------------------------------+
    | x != y    | Checks if two users are not equal.          |
    +-----------+---------------------------------------------+
    | hash(x)   | Return the user's hash.                     |
    +-----------+---------------------------------------------+
    | str(x)    | Returns the user's name with discriminator. |
    +-----------+---------------------------------------------+

    Attributes
    -----------
    name : str
        The user's username.
    id : str
        The user's unique ID.
    discriminator : str or int
        The user's discriminator. This is given when the username has conflicts.
    avatar : str
        The avatar hash the user has. Could be None.
    bot : bool
        Specifies if the user is a bot account.
    """

    __slots__ = ['name', 'id', 'discriminator', 'avatar', 'bot']

    def __init__(self, **kwargs):
        self.name = kwargs.get('username')
        self.id = kwargs.get('id')
        self.discriminator = kwargs.get('discriminator')
        self.avatar = kwargs.get('avatar')
        self.bot = kwargs.get('bot', False)

    def __str__(self):
        return '{0.name}#{0.discriminator}'.format(self)

    def __eq__(self, other):
        return isinstance(other, User) and other.id == self.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    @property
    def avatar_url(self):
        """Returns a friendly URL version of the avatar variable the user has. An empty string if
        the user has no avatar."""
        if self.avatar is None:
            return ''
        return 'https://discordapp.com/api/users/{0.id}/avatars/{0.avatar}.jpg'.format(self)

    @property
    def default_avatar(self):
        """Returns the default avatar for a given user. This is calculated by the user's descriminator"""
        return DefaultAvatar(int(self.discriminator) % len(DefaultAvatar))

    @property
    def default_avatar_url(self):
        """Returns a URL for a user's default avatar."""
        return 'https://discordapp.com/assets/{0.url}.png'.format(self.default_avatar)

    @property
    def mention(self):
        """Returns a string that allows you to mention the given user."""
        return '<@{0.id}>'.format(self)

    def permissions_in(self, channel):
        """An alias for :meth:`Channel.permissions_for`.

        Basically equivalent to:

        .. code-block:: python

            channel.permissions_for(self)

        Parameters
        -----------
        channel
            The channel to check your permissions for.
        """
        return channel.permissions_for(self)

    @property
    def created_at(self):
        """Returns the user's creation time in UTC.

        This is when the user's discord account was created."""
        return snowflake_time(self.id)

    @property
    def display_name(self):
        """Returns the user's display name.

        For regular users this is just their username, but
        if they have a server specific nickname then that
        is returned instead.
        """
        return getattr(self, 'nick', None) or self.name

    def mentioned_in(self, message):
        """Checks if the user is mentioned in the specified message.

        Parameters
        -----------
        message : :class:`Message`
            The message to check if you're mentioned in.
        """

        if message.mention_everyone:
            return True

        if self in message.mentions:
            return True

        return False



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import discord
import inspect
import importlib
import sys
import traceback
import re

from .core import GroupMixin, Command, command
from .view import StringView
from .context import Context
from .errors import CommandNotFound, CommandError
from .formatter import HelpFormatter

def _get_variable(name):
    stack = inspect.stack()
    try:
        for frames in stack:
            try:
                frame = frames[0]
                current_locals = frame.f_locals
                if name in current_locals:
                    return current_locals[name]
            finally:
                del frame
    finally:
        del stack

def when_mentioned(bot, msg):
    """A callable that implements a command prefix equivalent
    to being mentioned, e.g. ``@bot ``."""
    server = msg.server
    if server is not None:
        return '{0.me.mention} '.format(server)
    return '{0.user.mention} '.format(bot)

def when_mentioned_or(*prefixes):
    """A callable that implements when mentioned or other prefixes provided.

    Example
    --------

    .. code-block:: python

        bot = commands.Bot(command_prefix=commands.when_mentioned_or('!'))

    See Also
    ----------
    :func:`when_mentioned`
    """
    def inner(bot, msg):
        r = list(prefixes)
        r.append(when_mentioned(bot, msg))
        return r

    return inner

_mentions_transforms = {
    '@everyone': '@\u200beveryone',
    '@here': '@\u200bhere'
}

_mention_pattern = re.compile('|'.join(_mentions_transforms.keys()))

@asyncio.coroutine
def _default_help_command(ctx, *commands : str):
    """Shows this message."""
    bot = ctx.bot
    destination = ctx.message.author if bot.pm_help else ctx.message.channel

    def repl(obj):
        return _mentions_transforms.get(obj.group(0), '')

    # help by itself just lists our own commands.
    if len(commands) == 0:
        pages = bot.formatter.format_help_for(ctx, bot)
    elif len(commands) == 1:
        # try to see if it is a cog name
        name = _mention_pattern.sub(repl, commands[0])
        command = None
        if name in bot.cogs:
            command = bot.cogs[name]
        else:
            command = bot.commands.get(name)
            if command is None:
                yield from bot.send_message(destination, bot.command_not_found.format(name))
                return

        pages = bot.formatter.format_help_for(ctx, command)
    else:
        name = _mention_pattern.sub(repl, commands[0])
        command = bot.commands.get(name)
        if command is None:
            yield from bot.send_message(destination, bot.command_not_found.format(name))
            return

        for key in commands[1:]:
            try:
                key = _mention_pattern.sub(repl, key)
                command = command.commands.get(key)
                if command is None:
                    yield from bot.send_message(destination, bot.command_not_found.format(key))
                    return
            except AttributeError:
                yield from bot.send_message(destination, bot.command_has_no_subcommands.format(command, key))
                return

        pages = bot.formatter.format_help_for(ctx, command)

    if bot.pm_help is None:
        characters = sum(map(lambda l: len(l), pages))
        # modify destination based on length of pages.
        if characters > 1000:
            destination = ctx.message.author

    for page in pages:
        yield from bot.send_message(destination, page)


class Bot(GroupMixin, discord.Client):
    """Represents a discord bot.

    This class is a subclass of :class:`discord.Client` and as a result
    anything that you can do with a :class:`discord.Client` you can do with
    this bot.

    This class also subclasses :class:`GroupMixin` to provide the functionality
    to manage commands.

    Attributes
    -----------
    command_prefix
        The command prefix is what the message content must contain initially
        to have a command invoked. This prefix could either be a string to
        indicate what the prefix should be, or a callable that takes in the bot
        as its first parameter and :class:`discord.Message` as its second
        parameter and returns the prefix. This is to facilitate "dynamic"
        command prefixes. This callable can be either a regular function or
        a coroutine.

        The command prefix could also be a list or a tuple indicating that
        multiple checks for the prefix should be used and the first one to
        match will be the invocation prefix. You can get this prefix via
        :attr:`Context.prefix`.
    description : str
        The content prefixed into the default help message.
    self_bot : bool
        If ``True``, the bot will only listen to commands invoked by itself rather
        than ignoring itself. If ``False`` (the default) then the bot will ignore
        itself. This cannot be changed once initialised.
    formatter : :class:`HelpFormatter`
        The formatter used to format the help message. By default, it uses a
        the :class:`HelpFormatter`. Check it for more info on how to override it.
        If you want to change the help command completely (add aliases, etc) then
        a call to :meth:`remove_command` with 'help' as the argument would do the
        trick.
    pm_help : Optional[bool]
        A tribool that indicates if the help command should PM the user instead of
        sending it to the channel it received it from. If the boolean is set to
        ``True``, then all help output is PM'd. If ``False``, none of the help
        output is PM'd. If ``None``, then the bot will only PM when the help
        message becomes too long (dictated by more than 1000 characters).
        Defaults to ``False``.
    help_attrs : dict
        A dictionary of options to pass in for the construction of the help command.
        This allows you to change the command behaviour without actually changing
        the implementation of the command. The attributes will be the same as the
        ones passed in the :class:`Command` constructor. Note that ``pass_context``
        will always be set to ``True`` regardless of what you pass in.
    command_not_found : str
        The format string used when the help command is invoked with a command that
        is not found. Useful for i18n. Defaults to ``"No command called {} found."``.
        The only format argument is the name of the command passed.
    command_has_no_subcommands : str
        The format string used when the help command is invoked with requests for a
        subcommand but the command does not have any subcommands. Defaults to
        ``"Command {0.name} has no subcommands."``. The first format argument is the
        :class:`Command` attempted to get a subcommand and the second is the name.
    """
    def __init__(self, command_prefix, formatter=None, description=None, pm_help=False, **options):
        super().__init__(**options)
        self.command_prefix = command_prefix
        self.extra_events = {}
        self.cogs = {}
        self.extensions = {}
        self._checks = []
        self.description = inspect.cleandoc(description) if description else ''
        self.pm_help = pm_help
        self.command_not_found = options.pop('command_not_found', 'No command called "{}" found.')
        self.command_has_no_subcommands = options.pop('command_has_no_subcommands', 'Command {0.name} has no subcommands.')

        self._skip_check = discord.User.__ne__ if options.pop('self_bot', False) else discord.User.__eq__

        self.help_attrs = options.pop('help_attrs', {})
        self.help_attrs['pass_context'] = True

        if 'name' not in self.help_attrs:
            self.help_attrs['name'] = 'help'

        if formatter is not None:
            if not isinstance(formatter, HelpFormatter):
                raise discord.ClientException('Formatter must be a subclass of HelpFormatter')
            self.formatter = formatter
        else:
            self.formatter = HelpFormatter()

        # pay no mind to this ugliness.
        self.command(**self.help_attrs)(_default_help_command)

    # internal helpers

    @asyncio.coroutine
    def _get_prefix(self, message):
        prefix = self.command_prefix
        if callable(prefix):
            ret = prefix(self, message)
            if asyncio.iscoroutine(ret):
                ret = yield from ret
            return ret
        else:
            return prefix

    @asyncio.coroutine
    def _run_extra(self, coro, event_name, *args, **kwargs):
        try:
            yield from coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                yield from self.on_error(event_name, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def dispatch(self, event_name, *args, **kwargs):
        super().dispatch(event_name, *args, **kwargs)
        ev = 'on_' + event_name
        if ev in self.extra_events:
            for event in self.extra_events[ev]:
                coro = self._run_extra(event, event_name, *args, **kwargs)
                discord.compat.create_task(coro, loop=self.loop)

    @asyncio.coroutine
    def on_command_error(self, exception, context):
        """|coro|

        The default command error handler provided by the bot.

        By default this prints to ``sys.stderr`` however it could be
        overridden to have a different implementation.

        This only fires if you do not specify any listeners for command error.
        """
        if self.extra_events.get('on_command_error', None):
            return

        if hasattr(context.command, "on_error"):
            return

        print('Ignoring exception in command {}'.format(context.command), file=sys.stderr)
        traceback.print_exception(type(exception), exception, exception.__traceback__, file=sys.stderr)

    # utility "send_*" functions

    @asyncio.coroutine
    def _augmented_msg(self, coro, **kwargs):
        msg = yield from coro
        delete_after = kwargs.get('delete_after')
        if delete_after is not None:
            @asyncio.coroutine
            def delete():
                yield from asyncio.sleep(delete_after)
                yield from self.delete_message(msg)

            discord.compat.create_task(delete(), loop=self.loop)

        return msg

    def say(self, *args, **kwargs):
        """|coro|

        A helper function that is equivalent to doing

        .. code-block:: python

            self.send_message(message.channel, *args, **kwargs)

        The following keyword arguments are "extensions" that augment the
        behaviour of the standard wrapped call.

        Parameters
        ------------
        delete_after: float
            Number of seconds to wait before automatically deleting the
            message.

        See Also
        ---------
        :meth:`Client.send_message`
        """
        destination = _get_variable('_internal_channel')

        extensions = ('delete_after',)
        params = {
            k: kwargs.pop(k, None) for k in extensions
        }

        coro = self.send_message(destination, *args, **kwargs)
        return self._augmented_msg(coro, **params)

    def whisper(self, *args, **kwargs):
        """|coro|

        A helper function that is equivalent to doing

        .. code-block:: python

            self.send_message(message.author, *args, **kwargs)

        The following keyword arguments are "extensions" that augment the
        behaviour of the standard wrapped call.

        Parameters
        ------------
        delete_after: float
            Number of seconds to wait before automatically deleting the
            message.

        See Also
        ---------
        :meth:`Client.send_message`
        """
        destination = _get_variable('_internal_author')

        extensions = ('delete_after',)
        params = {
            k: kwargs.pop(k, None) for k in extensions
        }

        coro = self.send_message(destination, *args, **kwargs)
        return self._augmented_msg(coro, **params)

    def reply(self, content, *args, **kwargs):
        """|coro|

        A helper function that is equivalent to doing

        .. code-block:: python

            msg = '{0.mention}, {1}'.format(message.author, content)
            self.send_message(message.channel, msg, *args, **kwargs)

        The following keyword arguments are "extensions" that augment the
        behaviour of the standard wrapped call.

        Parameters
        ------------
        delete_after: float
            Number of seconds to wait before automatically deleting the
            message.

        See Also
        ---------
        :meth:`Client.send_message`
        """
        author = _get_variable('_internal_author')
        destination = _get_variable('_internal_channel')
        fmt = '{0.mention}, {1}'.format(author, str(content))

        extensions = ('delete_after',)
        params = {
            k: kwargs.pop(k, None) for k in extensions
        }

        coro = self.send_message(destination, fmt, *args, **kwargs)
        return self._augmented_msg(coro, **params)

    def upload(self, *args, **kwargs):
        """|coro|

        A helper function that is equivalent to doing

        .. code-block:: python

            self.send_file(message.channel, *args, **kwargs)

        The following keyword arguments are "extensions" that augment the
        behaviour of the standard wrapped call.

        Parameters
        ------------
        delete_after: float
            Number of seconds to wait before automatically deleting the
            message.

        See Also
        ---------
        :meth:`Client.send_file`
        """
        destination = _get_variable('_internal_channel')

        extensions = ('delete_after',)
        params = {
            k: kwargs.pop(k, None) for k in extensions
        }

        coro = self.send_file(destination, *args, **kwargs)
        return self._augmented_msg(coro, **params)

    def type(self):
        """|coro|

        A helper function that is equivalent to doing

        .. code-block:: python

            self.send_typing(message.channel)

        See Also
        ---------
        The :meth:`Client.send_typing` function.
        """
        destination = _get_variable('_internal_channel')
        return self.send_typing(destination)

    # global check registration

    def check(self, func):
        """A decorator that adds a global check to the bot.

        A global check is similar to a :func:`check` that is applied
        on a per command basis except it is run before any command checks
        have been verified and applies to every command the bot has.

        .. warning::

            This function must be a *regular* function and not a coroutine.

        Similar to a command :func:`check`\, this takes a single parameter
        of type :class:`Context` and can only raise exceptions derived from
        :exc:`CommandError`.

        Example
        ---------

        .. code-block:: python

            @bot.check
            def whitelist(ctx):
                return ctx.message.author.id in my_whitelist

        """
        self.add_check(func)
        return func

    def add_check(self, func):
        """Adds a global check to the bot.

        This is the non-decorator interface to :meth:`check`.

        Parameters
        -----------
        func
            The function that was used as a global check.
        """
        self._checks.append(func)

    def remove_check(self, func):
        """Removes a global check from the bot.

        This function is idempotent and will not raise an exception
        if the function is not in the global checks.

        Parameters
        -----------
        func
            The function to remove from the global checks.
        """

        try:
            self._checks.remove(func)
        except ValueError:
            pass

    def can_run(self, ctx):
        return all(f(ctx) for f in self._checks)

    # listener registration

    def add_listener(self, func, name=None):
        """The non decorator alternative to :meth:`listen`.

        Parameters
        -----------
        func : coroutine
            The extra event to listen to.
        name : Optional[str]
            The name of the command to use. Defaults to ``func.__name__``.

        Example
        --------

        .. code-block:: python

            async def on_ready(): pass
            async def my_message(message): pass

            bot.add_listener(on_ready)
            bot.add_listener(my_message, 'on_message')

        """
        name = func.__name__ if name is None else name

        if not asyncio.iscoroutinefunction(func):
            raise discord.ClientException('Listeners must be coroutines')

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def remove_listener(self, func, name=None):
        """Removes a listener from the pool of listeners.

        Parameters
        -----------
        func
            The function that was used as a listener to remove.
        name
            The name of the event we want to remove. Defaults to
            ``func.__name__``.
        """

        name = func.__name__ if name is None else name

        if name in self.extra_events:
            try:
                self.extra_events[name].remove(func)
            except ValueError:
                pass

    def listen(self, name=None):
        """A decorator that registers another function as an external
        event listener. Basically this allows you to listen to multiple
        events from different places e.g. such as :func:`discord.on_ready`

        The functions being listened to must be a coroutine.

        Example
        --------

        .. code-block:: python

            @bot.listen()
            async def on_message(message):
                print('one')

            # in some other file...

            @bot.listen('on_message')
            async def my_message(message):
                print('two')

        Would print one and two in an unspecified order.

        Raises
        -------
        discord.ClientException
            The function being listened to is not a coroutine.
        """

        def decorator(func):
            self.add_listener(func, name)
            return func

        return decorator

    # cogs

    def add_cog(self, cog):
        """Adds a "cog" to the bot.

        A cog is a class that has its own event listeners and commands.

        They are meant as a way to organize multiple relevant commands
        into a singular class that shares some state or no state at all.

        The cog can also have a ``__check`` member function that allows
        you to define a global check. See :meth:`check` for more info.

        More information will be documented soon.

        Parameters
        -----------
        cog
            The cog to register to the bot.
        """

        self.cogs[type(cog).__name__] = cog

        try:
            check = getattr(cog, '_{.__class__.__name__}__check'.format(cog))
        except AttributeError:
            pass
        else:
            self.add_check(check)

        members = inspect.getmembers(cog)
        for name, member in members:
            # register commands the cog has
            if isinstance(member, Command):
                member.instance = cog
                if member.parent is None:
                    self.add_command(member)
                continue

            # register event listeners the cog has
            if name.startswith('on_'):
                self.add_listener(member)

    def get_cog(self, name):
        """Gets the cog instance requested.

        If the cog is not found, ``None`` is returned instead.

        Parameters
        -----------
        name : str
            The name of the cog you are requesting.
        """
        return self.cogs.get(name)

    def remove_cog(self, name):
        """Removes a cog from the bot.

        All registered commands and event listeners that the
        cog has registered will be removed as well.

        If no cog is found then ``None`` is returned, otherwise
        the cog instance that is being removed is returned.

        If the cog defines a special member function named ``__unload``
        then it is called when removal has completed. This function
        **cannot** be a coroutine. It must be a regular function.

        Parameters
        -----------
        name : str
            The name of the cog to remove.
        """

        cog = self.cogs.pop(name, None)
        if cog is None:
            return cog

        members = inspect.getmembers(cog)
        for name, member in members:
            # remove commands the cog has
            if isinstance(member, Command):
                member.instance = None
                if member.parent is None:
                    self.remove_command(member.name)
                continue

            # remove event listeners the cog has
            if name.startswith('on_'):
                self.remove_listener(member)

        try:
            check = getattr(cog, '_{0.__class__.__name__}__check'.format(cog))
        except AttributeError:
            pass
        else:
            self.remove_check(check)

        unloader_name = '_{0.__class__.__name__}__unload'.format(cog)
        try:
            unloader = getattr(cog, unloader_name)
        except AttributeError:
            pass
        else:
            unloader()

        del cog

    # extensions

    def load_extension(self, name):
        if name in self.extensions:
            return

        lib = importlib.import_module(name)
        if not hasattr(lib, 'setup'):
            del lib
            del sys.modules[name]
            raise discord.ClientException('extension does not have a setup function')

        lib.setup(self)
        self.extensions[name] = lib

    def unload_extension(self, name):
        lib = self.extensions.get(name)
        if lib is None:
            return

        # find all references to the module

        # remove the cogs registered from the module
        for cogname, cog in self.cogs.copy().items():
            if inspect.getmodule(cog) is lib:
                self.remove_cog(cogname)

        # first remove all the commands from the module
        for command in self.commands.copy().values():
            if command.module is lib:
                command.module = None
                if isinstance(command, GroupMixin):
                    command.recursively_remove_all_commands()
                self.remove_command(command.name)

        # then remove all the listeners from the module
        for event_list in self.extra_events.copy().values():
            remove = []
            for index, event in enumerate(event_list):
                if inspect.getmodule(event) is lib:
                    remove.append(index)

            for index in reversed(remove):
                del event_list[index]

        # finally remove the import..
        del lib
        del self.extensions[name]
        del sys.modules[name]

    # command processing

    @asyncio.coroutine
    def process_commands(self, message):
        """|coro|

        This function processes the commands that have been registered
        to the bot and other groups. Without this coroutine, none of the
        commands will be triggered.

        By default, this coroutine is called inside the :func:`on_message`
        event. If you choose to override the :func:`on_message` event, then
        you should invoke this coroutine as well.

        Warning
        --------
        This function is necessary for :meth:`say`, :meth:`whisper`,
        :meth:`type`, :meth:`reply`, and :meth:`upload` to work due to the
        way they are written. It is also required for the :func:`on_command`
        and :func:`on_command_completion` events.

        Parameters
        -----------
        message : discord.Message
            The message to process commands for.
        """
        _internal_channel = message.channel
        _internal_author = message.author

        view = StringView(message.content)
        if self._skip_check(message.author, self.user):
            return

        prefix = yield from self._get_prefix(message)
        invoked_prefix = prefix

        if not isinstance(prefix, (tuple, list)):
            if not view.skip_string(prefix):
                return
        else:
            invoked_prefix = discord.utils.find(view.skip_string, prefix)
            if invoked_prefix is None:
                return


        invoker = view.get_word()
        tmp = {
            'bot': self,
            'invoked_with': invoker,
            'message': message,
            'view': view,
            'prefix': invoked_prefix
        }
        ctx = Context(**tmp)
        del tmp

        if invoker in self.commands:
            command = self.commands[invoker]
            self.dispatch('command', command, ctx)
            try:
                yield from command.invoke(ctx)
            except CommandError as e:
                ctx.command.dispatch_error(e, ctx)
            else:
                self.dispatch('command_completion', command, ctx)
        elif invoker:
            exc = CommandNotFound('Command "{}" is not found'.format(invoker))
            self.dispatch('command_error', exc, ctx)

    @asyncio.coroutine
    def on_message(self, message):
        yield from self.process_commands(message)



# -*- coding: utf-8 -*-

"""
discord.py extensions
~~~~~~~~~~~~~~~~~~~~~~

Extensions for the discord.py library live in this namespace.

:copyright: (c) 2016 Rapptz
:license: MIT, see LICENSE for more details.

"""



from setuptools import setup, find_packages
import re, os

on_rtd = os.getenv('READTHEDOCS') == 'True'

requirements = []
with open('requirements.txt') as f:
  requirements = f.read().splitlines()

if on_rtd:
  requirements.append('sphinxcontrib-napoleon')

version = ''
with open('discord/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('version is not set')

readme = ''
with open('README.md') as f:
    readme = f.read()

extras_require = {
    'voice': ['PyNaCl==1.0.1'],
}

setup(name='discord.py',
      author='Rapptz',
      url='https://github.com/Rapptz/discord.py',
      version=version,
      packages=find_packages(),
      license='MIT',
      description='A python wrapper for the Discord API',
      long_description=readme,
      include_package_data=True,
      install_requires=requirements,
      extras_require=extras_require,
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
      ]
)



# -*- coding: utf-8 -*-
"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import copy
from . import utils
from .permissions import Permissions, PermissionOverwrite
from .enums import ChannelType
from collections import namedtuple
from .mixins import Hashable
from .role import Role
from .user import User
from .member import Member

Overwrites = namedtuple('Overwrites', 'id allow deny type')

class Channel(Hashable):
    """Represents a Discord server channel.

    Supported Operations:

    +-----------+---------------------------------------+
    | Operation |              Description              |
    +===========+=======================================+
    | x == y    | Checks if two channels are equal.     |
    +-----------+---------------------------------------+
    | x != y    | Checks if two channels are not equal. |
    +-----------+---------------------------------------+
    | hash(x)   | Returns the channel's hash.           |
    +-----------+---------------------------------------+
    | str(x)    | Returns the channel's name.           |
    +-----------+---------------------------------------+

    Attributes
    -----------
    name : str
        The channel name.
    server : :class:`Server`
        The server the channel belongs to.
    id : str
        The channel ID.
    topic : Optional[str]
        The channel's topic. None if it doesn't exist.
    is_private : bool
        ``True`` if the channel is a private channel (i.e. PM). ``False`` in this case.
    position : int
        The position in the channel list. This is a number that starts at 0. e.g. the
        top channel is position 0. The position varies depending on being a voice channel
        or a text channel, so a 0 position voice channel is on top of the voice channel
        list.
    type : :class:`ChannelType`
        The channel type. There is a chance that the type will be ``str`` if
        the channel type is not within the ones recognised by the enumerator.
    bitrate : int
        The channel's preferred audio bitrate in bits per second.
    voice_members
        A list of :class:`Members` that are currently inside this voice channel.
        If :attr:`type` is not :attr:`ChannelType.voice` then this is always an empty array.
    user_limit : int
        The channel's limit for number of members that can be in a voice channel.
    """

    __slots__ = [ 'voice_members', 'name', 'id', 'server', 'topic', 'position',
                  'is_private', 'type', 'bitrate', 'user_limit',
                  '_permission_overwrites' ]

    def __init__(self, **kwargs):
        self._update(**kwargs)
        self.voice_members = []

    def __str__(self):
        return self.name

    def _update(self, **kwargs):
        self.name = kwargs.get('name')
        self.server = kwargs.get('server')
        self.id = kwargs.get('id')
        self.topic = kwargs.get('topic')
        self.is_private = False
        self.position = kwargs.get('position')
        self.bitrate = kwargs.get('bitrate')
        self.type = kwargs.get('type')
        self.user_limit = kwargs.get('user_limit')
        try:
            self.type = ChannelType(self.type)
        except:
            pass

        self._permission_overwrites = []
        everyone_index = 0
        everyone_id = self.server.id

        for index, overridden in enumerate(kwargs.get('permission_overwrites', [])):
            overridden_id = overridden['id']
            self._permission_overwrites.append(Overwrites(**overridden))

            if overridden.get('type') == 'member':
                continue

            if overridden_id == everyone_id:
                # the @everyone role is not guaranteed to be the first one
                # in the list of permission overwrites, however the permission
                # resolution code kind of requires that it is the first one in
                # the list since it is special. So we need the index so we can
                # swap it to be the first one.
                everyone_index = index

        # do the swap
        tmp = self._permission_overwrites
        if tmp:
            tmp[everyone_index], tmp[0] = tmp[0], tmp[everyone_index]

    @property
    def changed_roles(self):
        """Returns a list of :class:`Roles` that have been overridden from
        their default values in the :attr:`Server.roles` attribute."""
        ret = []
        for overwrite in filter(lambda o: o.type == 'role', self._permission_overwrites):
            role = utils.get(self.server.roles, id=overwrite.id)
            if role is None:
                continue

            role = copy.copy(role)
            role.permissions.handle_overwrite(overwrite.allow, overwrite.deny)
            ret.append(role)
        return ret

    @property
    def is_default(self):
        """bool : Indicates if this is the default channel for the :class:`Server` it belongs to."""
        return self.server.id == self.id

    @property
    def mention(self):
        """str : The string that allows you to mention the channel."""
        return '<#{0.id}>'.format(self)

    @property
    def created_at(self):
        """Returns the channel's creation time in UTC."""
        return utils.snowflake_time(self.id)

    def overwrites_for(self, obj):
        """Returns the channel-specific overwrites for a member or a role.

        Parameters
        -----------
        obj
            The :class:`Role` or :class:`Member` or :class:`Object` denoting
            whose overwrite to get.

        Returns
        ---------
        :class:`PermissionOverwrite`
            The permission overwrites for this object.
        """

        if isinstance(obj, Member):
            predicate = lambda p: p.type == 'member'
        elif isinstance(obj, Role):
            predicate = lambda p: p.type == 'role'
        else:
            predicate = lambda p: True

        for overwrite in filter(predicate, self._permission_overwrites):
            if overwrite.id == obj.id:
                allow = Permissions(overwrite.allow)
                deny = Permissions(overwrite.deny)
                return PermissionOverwrite.from_pair(allow, deny)

        return PermissionOverwrite()

    def permissions_for(self, member):
        """Handles permission resolution for the current :class:`Member`.

        This function takes into consideration the following cases:

        - Server owner
        - Server roles
        - Channel overrides
        - Member overrides
        - Whether the channel is the default channel.

        Parameters
        ----------
        member : :class:`Member`
            The member to resolve permissions for.

        Returns
        -------
        :class:`Permissions`
            The resolved permissions for the member.
        """

        # The current cases can be explained as:
        # Server owner get all permissions -- no questions asked. Otherwise...
        # The @everyone role gets the first application.
        # After that, the applied roles that the user has in the channel
        # (or otherwise) are then OR'd together.
        # After the role permissions are resolved, the member permissions
        # have to take into effect.
        # After all that is done.. you have to do the following:

        # If manage permissions is True, then all permissions are set to
        # True. If the channel is the default channel then everyone gets
        # read permissions regardless.

        # The operation first takes into consideration the denied
        # and then the allowed.

        if member.id == self.server.owner.id:
            return Permissions.all()

        default = self.server.default_role
        base = Permissions(default.permissions.value)

        # Apply server roles that the member has.
        for role in member.roles:
            base.value |= role.permissions.value

        # Server-wide Administrator -> True for everything
        # Bypass all channel-specific overrides
        if base.administrator:
            return Permissions.all()

        member_role_ids = set(map(lambda r: r.id, member.roles))
        denies = 0
        allows = 0

        # Apply channel specific role permission overwrites
        for overwrite in self._permission_overwrites:
            if overwrite.type == 'role' and overwrite.id in member_role_ids:
                denies |= overwrite.deny
                allows |= overwrite.allow

        base.handle_overwrite(allow=allows, deny=denies)

        # Apply member specific permission overwrites
        for overwrite in self._permission_overwrites:
            if overwrite.type == 'member' and overwrite.id == member.id:
                base.handle_overwrite(allow=overwrite.allow, deny=overwrite.deny)
                break

        # default channels can always be read
        if self.is_default:
            base.read_messages = True

        # if you can't send a message in a channel then you can't have certain
        # permissions as well
        if not base.send_messages:
            base.send_tts_messages = False
            base.mention_everyone = False
            base.embed_links = False
            base.attach_files = False

        # if you can't read a channel then you have no permissions there
        if not base.read_messages:
            denied = Permissions.all_channel()
            base.value &= ~denied.value

        # text channels do not have voice related permissions
        if self.type is ChannelType.text:
            denied = Permissions.voice()
            base.value &= ~denied.value

        return base

class PrivateChannel(Hashable):
    """Represents a Discord private channel.

    Supported Operations:

    +-----------+-------------------------------------------------+
    | Operation |                   Description                   |
    +===========+=================================================+
    | x == y    | Checks if two channels are equal.               |
    +-----------+-------------------------------------------------+
    | x != y    | Checks if two channels are not equal.           |
    +-----------+-------------------------------------------------+
    | hash(x)   | Returns the channel's hash.                     |
    +-----------+-------------------------------------------------+
    | str(x)    | Returns a string representation of the channel  |
    +-----------+-------------------------------------------------+

    Attributes
    ----------
    recipients: list of :class:`User`
        The users you are participating with in the private channel.
    me: :class:`User`
        The user presenting yourself.
    id: str
        The private channel ID.
    is_private: bool
        ``True`` if the channel is a private channel (i.e. PM). ``True`` in this case.
    type: :class:`ChannelType`
        The type of private channel.
    owner: Optional[:class:`User`]
        The user that owns the private channel. If the channel type is not
        :attr:`ChannelType.group` then this is always ``None``.
    icon: Optional[str]
        The private channel's icon hash. If the channel type is not
        :attr:`ChannelType.group` then this is always ``None``.
    name: Optional[str]
        The private channel's name. If the channel type is not
        :attr:`ChannelType.group` then this is always ``None``.
    """

    __slots__ = ['id', 'recipients', 'type', 'owner', 'icon', 'name', 'me']

    def __init__(self, me, **kwargs):
        self.recipients = [User(**u) for u in kwargs['recipients']]
        self.id = kwargs['id']
        self.me = me
        self.type = ChannelType(kwargs['type'])
        self._update_group(**kwargs)

    def _update_group(self, **kwargs):
        owner_id = kwargs.get('owner_id')
        self.icon = kwargs.get('icon')
        self.name = kwargs.get('name')
        self.owner = utils.find(lambda u: u.id == owner_id, self.recipients)

    @property
    def is_private(self):
        return True

    def __str__(self):
        if self.type is ChannelType.private:
            return 'Direct Message with {0.name}'.format(self.user)

        if self.name:
            return self.name

        if len(self.recipients) == 0:
            return 'Unnamed'

        return ', '.join(map(lambda x: x.name, self.recipients))

    @property
    def user(self):
        """A property that returns the first recipient of the private channel.

        This is mainly for compatibility and ease of use with old style private
        channels that had a single recipient.
        """
        return self.recipients[0]

    @property
    def icon_url(self):
        """Returns the channel's icon URL if available or an empty string otherwise."""
        if self.icon is None:
            return ''

        return 'https://cdn.discordapp.com/channel-icons/{0.id}/{0.icon}.jpg'.format(self)

    @property
    def created_at(self):
        """Returns the private channel's creation time in UTC."""
        return utils.snowflake_time(self.id)

    def permissions_for(self, user):
        """Handles permission resolution for a :class:`User`.

        This function is there for compatibility with :class:`Channel`.

        Actual private messages do not really have the concept of permissions.

        This returns all the Text related permissions set to true except:

        - send_tts_messages: You cannot send TTS messages in a PM.
        - manage_messages: You cannot delete others messages in a PM.

        This also handles permissions for :attr:`ChannelType.group` channels
        such as kicking or mentioning everyone.

        Parameters
        -----------
        user : :class:`User`
            The user to check permissions for.

        Returns
        --------
        :class:`Permissions`
            The resolved permissions for the user.
        """

        base = Permissions.text()
        base.send_tts_messages = False
        base.manage_messages = False
        base.mention_everyone = self.type is ChannelType.group

        if user == self.owner:
            base.kick_members = True

        return base





# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2016 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from . import utils
from .mixins import Hashable

class Emoji(Hashable):
    """Represents a custom emoji.

    Depending on the way this object was created, some of the attributes can
    have a value of ``None``.

    Supported Operations:

    +-----------+-----------------------------------------+
    | Operation |               Description               |
    +===========+=========================================+
    | x == y    | Checks if two emoji are the same.       |
    +-----------+-----------------------------------------+
    | x != y    | Checks if two emoji are not the same.   |
    +-----------+-----------------------------------------+
    | hash(x)   | Return the emoji's hash.                |
    +-----------+-----------------------------------------+
    | iter(x)   | Returns an iterator of (field, value)   |
    |           | pairs. This allows this class to be     |
    |           | used as an iterable in list/dict/etc.   |
    |           | constructions.                          |
    +-----------+-----------------------------------------+
    | str(x)    | Returns the emoji rendered for discord. |
    +-----------+-----------------------------------------+

    Attributes
    -----------
    name : str
        The name of the emoji.
    id : str
        The emoji's ID.
    require_colons : bool
        If colons are required to use this emoji in the client (:PJSalt: vs PJSalt).
    managed : bool
        If this emoji is managed by a Twitch integration.
    server : :class:`Server`
        The server the emoji belongs to.
    roles : List[:class:`Role`]
        A list of :class:`Role` that is allowed to use this emoji. If roles is empty,
        the emoji is unrestricted.
    """
    __slots__ = ["require_colons", "managed", "id", "name", "roles", 'server']

    def __init__(self, **kwargs):
        self.server = kwargs.pop('server')
        self._from_data(kwargs)

    def _from_data(self, emoji):
        self.require_colons = emoji.get('require_colons')
        self.managed = emoji.get('managed')
        self.id = emoji.get('id')
        self.name = emoji.get('name')
        self.roles = emoji.get('roles', [])
        if self.roles:
            roles = set(self.roles)
            self.roles = [role for role in self.server.roles if role.id in roles]

    def _iterator(self):
        for attr in self.__slots__:
            value = getattr(self, attr, None)
            if value is not None:
                yield (attr, value)

    def __iter__(self):
        return self._iterator()

    def __str__(self):
        return "<:{0.name}:{0.id}>".format(self)

    @property
    def created_at(self):
        """Returns the emoji's creation time in UTC."""
        return utils.snowflake_time(self.id)

    @property
    def url(self):
        """Returns a URL version of the emoji."""
        return "https://discordapp.com/api/emojis/{0.id}.png".format(self)
